import argparse
import pytest
import os
import sys
import os
import textwrap
import yaml
import mock
import glob
import socket
import six
import time
import json
import threading
from click.testing import CliRunner

from .api_mocks import *
from .utils import runner, git_repo
import wandb
from wandb import wandb_run
from wandb import env

# Tests which rely on row history in memory should set `History.keep_rows = True`
from wandb.history import History
History.keep_rows = True


def test_log(wandb_init_run):
    history_row = {'stuff': 5}
    wandb.log(history_row)
    assert len(wandb.run.history.rows) == 1
    assert set(history_row.items()) <= set(wandb.run.history.rows[0].items())


@pytest.mark.mocked_run_manager()
def test_log_step(wandb_init_run, capsys):
    history_row = {'stuff': 5}
    wandb.log(history_row, step=5)
    wandb.log()
    _, err = capsys.readouterr()
    assert "wandb: " in err
    assert len(wandb.run.history.rows) == 1
    assert wandb.run.history.rows[0]['_step'] == 5


@pytest.mark.silent()
@pytest.mark.mocked_run_manager()
def test_log_silent(wandb_init_run, capsys):
    wandb.log({"cool": 1})
    _, err = capsys.readouterr()
    assert "wandb: " not in err


def test_log_only_strings_as_keys(wandb_init_run):
    with pytest.raises(ValueError):
        wandb.log({1: 1000})
    with pytest.raises(ValueError):
        wandb.log({('tup', 'idx'): 1000})


def test_async_log(wandb_init_run):
    for i in range(100):
        wandb.log({"cool": 1000}, sync=False)
    wandb.shutdown_async_log_thread()
    wandb.log({"cool": 100}, sync=False)
    wandb.shutdown_async_log_thread()
    assert wandb.run.history.rows[-1]['cool'] == 100
    assert len(wandb.run.history.rows) == 101


def test_nice_log_error():
    with pytest.raises(ValueError):
        wandb.log({"no": "init"})


def test_nice_log_error_config():
    with pytest.raises(wandb.Error) as e:
        wandb.config.update({"foo": 1})
    assert e.value.message == "You must call wandb.init() before wandb.config.update"
    with pytest.raises(wandb.Error) as e:
        wandb.config.foo = 1
    assert e.value.message == "You must call wandb.init() before wandb.config.foo"


def test_nice_log_error_summary():
    with pytest.raises(wandb.Error) as e:
        wandb.summary["great"] = 1
    assert e.value.message == 'You must call wandb.init() before wandb.summary["great"]'
    with pytest.raises(wandb.Error) as e:
        wandb.summary.bam = 1
    assert e.value.message == 'You must call wandb.init() before wandb.summary.bam'


@pytest.mark.args(k8s=True)
def test_k8s_success(wandb_init_run):
    assert os.getenv("WANDB_DOCKER") == "test@sha256:1234"


@pytest.mark.args(k8s=False)
def test_k8s_failure(wandb_init_run):
    assert os.getenv("WANDB_DOCKER") is None


@pytest.mark.args(sagemaker=True)
def test_sagemaker(wandb_init_run):
    assert wandb.config.fuckin == "A"
    assert wandb.run.id == "sage-maker"
    assert os.getenv('WANDB_TEST_SECRET') == "TRUE"
    assert wandb.run.group == "sage"


@pytest.mark.args(tf_config={"cluster": {"master": ["trainer-4dsl7-master-0:2222"]}, "task": {"type": "master", "index": 0}, "environment": "cloud"})
def test_simple_tfjob(wandb_init_run):
    assert wandb.run.group is None
    assert wandb.run.job_type == "master"


@pytest.mark.args(tf_config={"cluster": {"master": ["trainer-sj2hp-master-0:2222"], "ps": ["trainer-sj2hp-ps-0:2222"], "worker": ["trainer-sj2hp-worker-0:2222"]}, "task": {"type": "worker", "index": 0}, "environment": "cloud"})
def test_distributed_tfjob(wandb_init_run):
    assert wandb.run.group == "trainer-sj2hp"
    assert wandb.run.job_type == "worker"


@pytest.mark.args(tf_config={"cluster": {"corrupt": ["bad"]}})
def test_corrupt_tfjob(wandb_init_run):
    assert wandb.run.group is None


@pytest.mark.args(env={"TF_CONFIG": "garbage"})
def test_bad_json_tfjob(wandb_init_run):
    assert wandb.run.group is None


@pytest.mark.args(error="io")
def test_io_error(wandb_init_run, capsys):
    out, err = capsys.readouterr()
    assert isinstance(wandb_init_run, wandb.LaunchError)


@pytest.mark.headless()
@pytest.mark.args(error="socket")
def test_io_headless(wandb_init_run, mocker):
    with pytest.raises(wandb.LaunchError) as err:
        wandb._init_headless(wandb_init_run)
    assert "wandb/debug.log" in str(err.value)


@pytest.mark.args(dir="/tmp")
def test_custom_dir(wandb_init_run):
    assert len(glob.glob("/tmp/wandb/run-*")) > 0


def test_login_key(local_netrc, capsys):
    wandb.login(key="A"* 40)
    out, err = capsys.readouterr()
    assert "wandb: WARNING If" in err
    assert wandb.api.api_key == "A" * 40


def test_login_existing_key(local_netrc):
    os.environ["WANDB_API_KEY"] = "B" * 40
    wandb.ensure_configured()
    wandb.login()
    assert wandb.api.api_key == "B" * 40


def test_login_no_key(local_netrc, mocker):
    stdin_mock = mocker.patch("wandb.util.sys.stdin.isatty")
    stdin_mock.return_value = True
    stdout_mock = mocker.patch("wandb.util.sys.stdout.isatty")
    stdout_mock.return_value = True
    inp = mocker.patch("wandb.util.six.moves.input")
    inp.return_value = "2"
    getpass = mocker.patch("wandb.util.getpass.getpass")
    getpass.return_value = "C" * 40

    wandb.ensure_configured()
    assert wandb.api.api_key == None
    wandb.login()
    assert wandb.api.api_key == "C" * 40


def test_run_context_multi_run(live_mock_server, git_repo):
    os.environ[env.BASE_URL] = "http://localhost:%i" % 8765
    os.environ["WANDB_API_KEY"] = "B" * 40
    with wandb.init() as run:
        run.log({"a": 1, "b": 2})

    with wandb.init(reinit=True) as run:
        run.log({"c": 3, "d": 4})

    assert len(glob.glob("wandb/*")) == 4


def test_login_jupyter_anonymous(mock_server, local_netrc, mocker):
    python = mocker.patch("wandb._get_python_type")
    python.return_value = "ipython"
    wandb.login(anonymous="allow")
    assert wandb.api.api_key == "ANONYMOOSE" * 4


def test_login_anonymous(mock_server, local_netrc):
    os.environ["WANDB_API_KEY"] = "B" * 40
    wandb.login(anonymous="must")
    assert wandb.api.api_key == "ANONYMOOSE" * 4


@pytest.mark.mock_socket
def test_save_policy_symlink(wandb_init_run):
    with open("test.rad", "w") as f:
        f.write("something")
    wandb.save("test.rad")
    assert wandb_init_run.socket.send.called


@pytest.mark.mock_socket
def test_save_absolute_path(wandb_init_run):
    with open("/tmp/test.txt", "w") as f:
        f.write("something")
    wandb.save("/tmp/test.txt")
    assert os.path.exists(os.path.join(wandb_init_run.dir, "test.txt"))


@pytest.mark.mock_socket
def test_save_relative_path(wandb_init_run):
    with open("/tmp/test.txt", "w") as f:
        f.write("something")
    wandb.save("/tmp/test.txt", base_path="/")
    assert os.path.exists(os.path.join(wandb_init_run.dir, "tmp/test.txt"))


@pytest.mark.mock_socket
def test_save_invalid_path(wandb_init_run):
    with open("/tmp/test.txt", "w") as f:
        f.write("something")
    with pytest.raises(ValueError):
        wandb.save("../tmp/../../*.txt", base_path="/tmp")


@pytest.mark.args(resume=True)
def test_auto_resume_first(wandb_init_run):
    assert json.load(open(os.path.join(wandb.wandb_dir(), wandb_run.RESUME_FNAME)))[
        "run_id"] == wandb_init_run.id
    assert not wandb_init_run.resumed


@pytest.mark.args(resume="testy")
def test_auto_resume_manual(wandb_init_run):
    assert wandb_init_run.id == "testy"


@pytest.mark.resume()
@pytest.mark.args(resume=True)
def test_auto_resume_second(wandb_init_run):
    assert wandb_init_run.id == "test"
    assert wandb_init_run.resumed
    assert wandb_init_run.step == 16


@pytest.mark.resume()
@pytest.mark.args(resume=False)
def test_auto_resume_remove(wandb_init_run):
    assert not os.path.exists(os.path.join(
        wandb.wandb_dir(), wandb_run.RESUME_FNAME))


@pytest.mark.jupyter
def test_save_policy_jupyter(wandb_init_run, query_upload_h5, request_mocker):
    with open("test.rad", "w") as f:
        f.write("something")
    #mock = query_upload_h5(request_mocker)
    wandb.run.socket = None
    wandb.save("test.rad")
    assert wandb_init_run._jupyter_agent.rm._user_file_policies == {
        'end': [], 'live': ['test.rad']}


def test_restore(runner, request_mocker, download_url, wandb_init_run):
    with runner.isolated_filesystem():
        download_url(request_mocker, size=10000)
        res = wandb.restore("weights.h5")
        assert os.path.getsize(res.name) == 10000


@pytest.mark.jupyter
@pytest.mark.mocked_run_manager
def test_jupyter_init(wandb_init_run):
    assert os.getenv(env.JUPYTER)
    wandb.log({"stat": 1})
    fsapi = wandb_init_run.run_manager._api._file_stream_api
    wandb_init_run._stop_jupyter_agent()
    payloads = {c[1][0]: json.loads(c[1][1])
                for c in fsapi.push.mock_calls}
    assert payloads["wandb-history.jsonl"]["stat"] == 1
    assert payloads["wandb-history.jsonl"]["_step"] == 16

    # TODO: saw some global state issues here...
    # assert "" == err


@pytest.mark.skip
@pytest.mark.jupyter
def test_jupyter_log_history(wandb_init_run, capsys):
    # This simulates what the happens in a Jupyter notebook, it's gnarly
    # because it resumes so this depends on the run_resume_status which returns
    # a run that's at step 15 so calling log will update step to 16
    wandb.log({"something": "new"})
    rm = wandb_init_run.run_manager
    fsapi = rm._api._file_stream_api
    wandb_init_run._stop_jupyter_agent()
    files = [c[1][0] for c in fsapi.push.mock_calls]
    assert sorted(files) == ['wandb-events.jsonl',
                             'wandb-history.jsonl', 'wandb-summary.json']
    # TODO: There's a race here where a thread isn't stopped
    time.sleep(1)
    wandb.log({"resumed": "log"})
    new_fsapi = wandb_init_run._jupyter_agent.rm._api.get_file_stream_api()
    wandb_init_run.run_manager.test_shutdown()
    payloads = {c[1][0]: json.loads(c[1][1])
                for c in new_fsapi.push.mock_calls}
    assert payloads["wandb-history.jsonl"]["_step"] == 16
    assert payloads["wandb-history.jsonl"]["resumed"] == "log"


@pytest.mark.args(tensorboard=True)
@pytest.mark.skipif(sys.version_info < (3, 6) or os.environ.get("NO_ML") == "true", reason="no tensorboardX in py2 or no ml tests")
def test_tensorboard(wandb_init_run):
    from tensorboardX import SummaryWriter
    writer = SummaryWriter()
    writer.add_scalar('foo', 1, 0)
    writer.close()
    print("Real run: %s", wandb.run)
    print(wandb.run.history.row)
    print(wandb.run.history.rows)
    assert wandb.run.history.row['global_step'] == 0
    assert wandb.run.history.row['foo'] == 1.0


@pytest.mark.args(id="123456")
def test_run_id(wandb_init_run):
    assert wandb.run.id == "123456"


@pytest.mark.args(name="coolio")
def test_run_name(wandb_init_run):
    assert wandb.run.name == "coolio"


@pytest.mark.unconfigured
def test_not_logged_in(wandb_init_run, capsys):
    out, err = capsys.readouterr()
    assert "No credentials found.  Run \"wandb login\" to visualize your metrics" in err
    assert "_init_headless called with cloud=False" in out


@pytest.mark.jupyter
@pytest.mark.unconfigured
def test_jupyter_manual_configure(wandb_init_run, capsys):
    out, err = capsys.readouterr()
    assert "Not authenticated" in err
    assert "Python.core.display.HTML" in out



from __future__ import absolute_import
import sys
import os
# Ensure we don't have the wandb directory in the path to avoid importing our tensorboard
# module.  This should only happen when wandb is installed with pip -e or pip install ...#egg=wandb
for path in sys.path:
    if path.endswith(os.path.join("client", "wandb")):
        sys.path.remove(path)
    if path.endswith(os.path.join("site-packages", "wandb")):
        sys.path.remove(path)
if sys.modules.get("tensorboard"):
    del sys.modules["tensorboard"]
from tensorboard.backend.event_processing import directory_watcher
from tensorboard.backend.event_processing import event_file_loader
from tensorboard.compat import tf
from wandb.tensorboard import log
from wandb import util
import six
import os
from six.moves import queue
import wandb
import time
import threading
import collections


class Event(object):
    """An event wrapper to enable priority queueing"""

    def __init__(self, event, namespace):
        self.event = event
        self.namespace = namespace
        self.created_at = time.time()

    def __lt__(self, other):
        return self.event.wall_time < other.event.wall_time


class Consumer(object):
    """Consumes tfevents from a priority queue.  There should always
    only be one of these per run_manager.  We wait for 10 seconds of queued
    events to reduce the chance of multiple tfevent files triggering
    out of order steps.
    """

    def __init__(self, queue, delay=10):
        self._queue = queue
        self._thread = threading.Thread(target=self._thread_body)
        self._thread.daemon = True
        self._shutdown = False
        self._delay = delay

    def start(self):
        self._thread.start()

    def shutdown(self):
        self._delay = 0
        self._shutdown = True
        try:
            self._thread.join()
        # Incase we never start it
        except RuntimeError:
            pass

    def _thread_body(self):
        while True:
            try:
                event = self._queue.get(True, 1)
                # If the event was added later than delay, put it back in the queue
                if event.created_at > time.time() - self._delay:
                    self._queue.put(event)
                    time.sleep(0.1)
            except queue.Empty:
                event = None
                if self._shutdown:
                    break
            if event:
                self._handle_event(event)

    def _handle_event(self, event):
        log(event.event, step=event.event.step, namespace=event.namespace)


def loader(save=True, namespace=None):
    """Incredibly hacky class generator to optionally save / prefix tfevent files"""
    class EventFileLoader(event_file_loader.EventFileLoader):
        def __init__(self, file_path):
            super(EventFileLoader, self).__init__(file_path)
            if save:
                # TODO: save plugins?
                logdir = os.path.dirname(file_path)
                parts = list(os.path.split(logdir))
                if namespace and parts[-1] == namespace:
                    parts.pop()
                    logdir = os.path.join(*parts)
                wandb.save(file_path, base_path=logdir)
    return EventFileLoader


def IsNewTensorFlowEventsFile(path):
    """Checks if a path has been modified since launch and contains tfevents"""
    if not path:
        raise ValueError('Path must be a nonempty string')
    path = tf.compat.as_str_any(path)
    return 'tfevents' in os.path.basename(path) and os.stat(path).st_mtime >= wandb.START_TIME


class Watcher(object):
    def __init__(self, logdir, queue, namespace=None, save=True):
        """Uses tensorboard to watch a directory for tfevents files 
        and put them on a queue"""
        self.namespace = namespace
        self.queue = queue
        self.logdir = logdir
        # TODO: prepend the namespace here?
        self._generator = directory_watcher.DirectoryWatcher(
            logdir,
            loader(save, namespace),
            IsNewTensorFlowEventsFile)
        self._first_event_timestamp = None
        self._shutdown = False
        self._thread = threading.Thread(target=self._thread_body)
        self._thread.daemon = True

    def start(self):
        self._thread.start()

    def shutdown(self):
        self._shutdown = True
        try:
            self._thread.join()
        # Incase we never start it
        except RuntimeError:
            pass

    def _thread_body(self):
        """Check for new events every second"""
        while True:
            try:
                for event in self._generator.Load():
                    self.process_event(event)
            except directory_watcher.DirectoryDeletedError:
                break
            if self._shutdown:
                break
            else:
                time.sleep(1)

    def process_event(self, event):
        if self._first_event_timestamp is None:
            self._first_event_timestamp = event.wall_time

        if event.HasField('file_version'):
            self.file_version = event.file_version

        if event.HasField('summary'):
            self.queue.put(Event(event, self.namespace))



import numpy as np
import pandas as pd
import matplotlib.pyplot as mplt
import os
import collections
import sklearn
import scipy
from sklearn.base import BaseEstimator
from sklearn.model_selection import train_test_split
from sklearn.decomposition import PCA
from sklearn.neighbors import NearestNeighbors
from sklearn.neighbors import KNeighborsClassifier
import random
import wandb
from scipy.spatial.distance import euclidean, squareform, pdist
from sklearn.model_selection import GridSearchCV
from sklearn.svm import SVC
from sklearn.metrics import accuracy_score, f1_score
from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted
from sklearn.preprocessing import LabelEncoder

dimensionality_reduction=PCA(n_components=2)
def encode_labels(df):
    le = LabelEncoder()
    # apply le on categorical feature columns
    categorical_cols = df.select_dtypes(exclude=['int','float','float64','float32','int32','int64']).columns
    df[categorical_cols] = df[categorical_cols].apply(lambda col: le.fit_transform(col))
# Test Asummptions for plotting parameters and datasets
def test_missing(**kwargs):
    test_passed = True
    for k,v in kwargs.items():
        # Missing/empty params/datapoint arrays
        if v is None:
            wandb.termerror("%s is None. Please try again." % (k))
            test_passed = False
        if ((k == 'X') or (k == 'X_test')):
            if isinstance(v, scipy.sparse.csr.csr_matrix):
                v = v.toarray()
            elif isinstance(v, (pd.DataFrame, pd.Series)):
                v = v.to_numpy()
            elif isinstance(v, list):
                v = np.asarray(v)

            # Warn the user about missing values
            missing = 0
            missing = np.count_nonzero(pd.isnull(v))
            if missing>0:
                wandb.termwarn("%s contains %d missing values. " % (k,missing))
                test_passed = False
            # Ensure the dataset contains only integers
            non_nums = 0
            if v.ndim == 1:
                non_nums = sum(1 for val in v if (not isinstance(val, (int, float, complex)) and not isinstance(val,np.number)))
            else:
                non_nums = sum(1 for sl in v for val in sl if (not isinstance(val, (int, float, complex)) and not isinstance(val,np.number)))
            if non_nums>0:
                wandb.termerror("%s contains values that are not numbers. Please vectorize, label encode or one hot encode %s and call the plotting function again." % (k,k))
                test_passed = False
    return test_passed
def test_types(**kwargs):
    test_passed = True
    for k,v in kwargs.items():
        # check for incorrect types
        if ((k == 'X') or (k == 'X_test') or (k == 'y') or (k == 'y_test')
            or (k == 'y_true') or (k == 'y_probas')):
            if not isinstance(v, (collections.Sequence, collections.Iterable, np.ndarray, np.generic, pd.DataFrame, pd.Series, list)):
                wandb.termerror("%s is not an array. Please try again." % (k))
                test_passed = False
        # check for classifier types
        if (k=='model'):
            if ((not sklearn.base.is_classifier(v)) and (not sklearn.base.is_regressor(v))):
                wandb.termerror("%s is not a classifier or regressor. Please try again." % (k))
                test_passed = False
        elif (k=='clf' or k=='binary_clf'):
            if (not(sklearn.base.is_classifier(v))):
                wandb.termerror("%s is not a classifier. Please try again." % (k))
                test_passed = False
        elif (k=='regressor'):
            if (not sklearn.base.is_regressor(v)):
                wandb.termerror("%s is not a regressor. Please try again." % (k))
                test_passed = False
        elif (k=='clusterer'):
            if (not(getattr(v, "_estimator_type", None) == "clusterer")):
                wandb.termerror("%s is not a clusterer. Please try again." % (k))
                test_passed = False
    return test_passed
def test_fitted(model):
    try:
        model.predict(np.zeros((7, 3)))
    except NotFittedError:
        wandb.termerror("Please fit the model before passing it in.")
        return False
    except AttributeError:
        # Some clustering models (LDA, PCA, Agglomerative) don't implement ``predict``
        try:
            check_is_fitted(
                model,
                [
                    "coef_",
                    "estimator_",
                    "labels_",
                    "n_clusters_",
                    "children_",
                    "components_",
                    "n_components_",
                    "n_iter_",
                    "n_batch_iter_",
                    "explained_variance_",
                    "singular_values_",
                    "mean_",
                ],
                all_or_any=any,
            )
            return True
        except sklearn.exceptions.NotFittedError:
            wandb.termerror("Please fit the model before passing it in.")
            return False
    except Exception:
        # Assume it's fitted, since ``NotFittedError`` wasn't raised
        return True

# Decision Boundary Utils
# Modified from https://github.com/tmadl/highdimensional-decision-boundary-plot
class DBPlot(BaseEstimator):
    def __init__(
        self,
        estimator=KNeighborsClassifier(n_neighbors=10),
        acceptance_threshold=0.03,
        n_decision_boundary_keypoints=60,
        n_connecting_keypoints=None,
        n_interpolated_keypoints=None,
        n_generated_testpoints_per_keypoint=15,
        linear_iteration_budget=100,
        hypersphere_iteration_budget=300,
        verbose=False,
    ):
        if acceptance_threshold == 0:
            raise Warning(
                "A nonzero acceptance threshold is strongly recommended so the optimizer can finish in finite time"
            )
        if linear_iteration_budget < 2 or hypersphere_iteration_budget < 2:
            raise Exception("Invalid iteration budget")

        self.classifier = estimator
        self.acceptance_threshold = acceptance_threshold

        if (
            n_decision_boundary_keypoints
            and n_connecting_keypoints
            and n_interpolated_keypoints
            and n_connecting_keypoints + n_interpolated_keypoints
            != n_decision_boundary_keypoints
        ):
            raise Exception(
                "n_connecting_keypoints and n_interpolated_keypoints must sum to n_decision_boundary_keypoints (set them to None to use calculated suggestions)"
            )

        self.n_connecting_keypoints = (
            n_connecting_keypoints
            if n_connecting_keypoints != None
            else n_decision_boundary_keypoints / 3
        )
        self.n_interpolated_keypoints = (
            n_interpolated_keypoints
            if n_interpolated_keypoints != None
            else n_decision_boundary_keypoints * 2 / 3
        )

        self.linear_iteration_budget = linear_iteration_budget
        self.n_generated_testpoints_per_keypoint = n_generated_testpoints_per_keypoint
        self.hypersphere_iteration_budget = hypersphere_iteration_budget
        self.verbose = verbose

        self.decision_boundary_points = []
        self.decision_boundary_points_2d = []
        self.X_testpoints = []
        self.y_testpoints = []
        self.train_idx = []
        self.test_idx = []
        self.background = []
        self.steps = 3

        self.hypersphere_max_retry_budget = 20
        self.penalties_enabled = True
        self.random_gap_selection = False

    def setclassifier(self, estimator=KNeighborsClassifier(n_neighbors=10)):
        self.classifier = estimator

    def fit(self, X, y, training_indices=None):
        if set(np.array(y, dtype=int).tolist()) != set([0, 1]):
            raise Exception(
                "Currently only implemented for binary classification. Make sure you pass in two classes (0 and 1)"
            )

        if training_indices == None:
            train_idx = range(len(y))
        else:
            train_idx = training_indices

        self.X = X
        self.y = y
        self.train_idx = train_idx
        # self.test_idx = np.setdiff1d(np.arange(len(y)), self.train_idx, assume_unique=False)
        self.test_idx = list(set(range(len(y))).difference(set(self.train_idx)))

        # fit classifier if necessary
        try:
            self.classifier.predict([X[0]])
        except:
            self.classifier.fit(X, y)

        self.y_pred = self.classifier.predict(self.X)

        # fit DR method if necessary
        dimensionality_reduction.fit(X)
        try:
            dimensionality_reduction.transform([X[0]])
        except:
            raise Exception(
                "Please make sure your dimensionality reduction method has an exposed transform() method! If in doubt, use PCA or Isomap"
            )

        # transform data
        self.X2d = dimensionality_reduction.transform(self.X)
        self.mean_2d_dist = np.mean(pdist(self.X2d))
        self.X2d_xmin, self.X2d_xmax = np.min(self.X2d[:, 0]), np.max(self.X2d[:, 0])
        self.X2d_ymin, self.X2d_ymax = np.min(self.X2d[:, 1]), np.max(self.X2d[:, 1])

        self.majorityclass = 0 if list(y).count(0) > list(y).count(1) else 1
        self.minorityclass = 1 - self.majorityclass
        minority_idx, majority_idx = (
            np.where(y == self.minorityclass)[0],
            np.where(y == self.majorityclass)[0],
        )
        self.Xminor, self.Xmajor = X[minority_idx], X[majority_idx]
        self.Xminor2d, self.Xmajor2d = self.X2d[minority_idx], self.X2d[majority_idx]

        # set up efficient nearest neighbor models for later use
        self.nn_model_2d_majorityclass = NearestNeighbors(n_neighbors=2)
        self.nn_model_2d_majorityclass.fit(self.X2d[majority_idx, :])

        self.nn_model_2d_minorityclass = NearestNeighbors(n_neighbors=2)
        self.nn_model_2d_minorityclass.fit(self.X2d[minority_idx, :])

        # step 1. look for decision boundary points between corners of majority &
        # minority class distribution
        minority_corner_idx, majority_corner_idx = [], []
        for extremum1 in [np.min, np.max]:
            for extremum2 in [np.min, np.max]:
                _, idx = self.nn_model_2d_minorityclass.kneighbors(
                    [[extremum1(self.Xminor2d[:, 0]), extremum2(self.Xminor2d[:, 1])]]
                )
                minority_corner_idx.append(idx[0][0])
                _, idx = self.nn_model_2d_majorityclass.kneighbors(
                    [[extremum1(self.Xmajor2d[:, 0]), extremum2(self.Xmajor2d[:, 1])]]
                )
                majority_corner_idx.append(idx[0][0])

        # optimize to find new db keypoints between corners
        self._linear_decision_boundary_optimization(
            minority_corner_idx, majority_corner_idx, all_combinations=True, step=1
        )

        # step 2. look for decision boundary points on lines connecting randomly
        # sampled points of majority & minority class
        n_samples = int(self.n_connecting_keypoints)
        from_idx = list(random.sample(list(np.arange(len(self.Xminor))), n_samples))
        to_idx = list(random.sample(list(np.arange(len(self.Xmajor))), n_samples))

        # optimize to find new db keypoints between minority and majority class
        self._linear_decision_boundary_optimization(
            from_idx, to_idx, all_combinations=False, step=2
        )

        if len(self.decision_boundary_points_2d) < 2:
            wandb.termerror(
                "Failed to find initial decision boundary. Retrying... If this keeps happening, increasing the acceptance threshold might help. Also, make sure the classifier is able to find a point with 0.5 prediction probability (usually requires an even number of estimators/neighbors/etc)."
            )
            return self.fit(X, y, training_indices)

        # step 3. look for decision boundary points between already known db
        # points that are too distant (search on connecting line first, then on
        # surrounding hypersphere surfaces)
        edges, gap_distances, gap_probability_scores = (
            self._get_sorted_db_keypoint_distances()
        )  # find gaps
        self.nn_model_decision_boundary_points = NearestNeighbors(n_neighbors=2)
        self.nn_model_decision_boundary_points.fit(self.decision_boundary_points)

        i = 0
        retries = 0
        while i < self.n_interpolated_keypoints:
            if self.random_gap_selection:
                # randomly sample from sorted DB keypoint gaps?
                gap_idx = np.random.choice(
                    len(gap_probability_scores), 1, p=gap_probability_scores
                )[0]
            else:
                # get largest gap
                gap_idx = 0
            from_point = self.decision_boundary_points[edges[gap_idx][0]]
            to_point = self.decision_boundary_points[edges[gap_idx][1]]

            # optimize to find new db keypoint along line connecting two db keypoints
            # with large gap
            db_point = self._find_decision_boundary_along_line(
                from_point, to_point, penalize_tangent_distance=self.penalties_enabled
            )

            if self.decision_boundary_distance(db_point) > self.acceptance_threshold:
                if self.verbose:
                    wandb.termerror(
                        "No good solution along straight line - trying to find decision boundary on hypersphere surface around known decision boundary point"
                    )

                # hypersphere radius half the distance between from and to db keypoints
                R = euclidean(from_point, to_point) / 2.0
                # search around either source or target keypoint, with 0.5 probability,
                # hoping to find decision boundary in between
                if random.random() > 0.5:
                    from_point = to_point

                # optimize to find new db keypoint on hypersphere surphase around known keypoint
                db_point = self._find_decision_boundary_on_hypersphere(from_point, R)
                if (
                    self.decision_boundary_distance(db_point)
                    <= self.acceptance_threshold
                ):
                    db_point2d = dimensionality_reduction.transform([db_point])[0]
                    self.decision_boundary_points.append(db_point)
                    self.decision_boundary_points_2d.append(db_point2d)
                    i += 1
                    retries = 0
                else:
                    retries += 1
                    if retries > self.hypersphere_max_retry_budget:
                        i += 1
                        dist = self.decision_boundary_distance(db_point)
                        msg = "Found point is too distant from decision boundary ({}), but retry budget exceeded ({})"
                        wandb.termerror(msg.format(dist, self.hypersphere_max_retry_budget))
                    elif self.verbose:
                        dist = self.decision_boundary_distance(db_point)
                        wandb.termerror(
                            "Found point is too distant from decision boundary ({}) retrying...".format(
                                dist
                            )
                        )

            else:
                db_point2d = dimensionality_reduction.transform([db_point])[0]
                self.decision_boundary_points.append(db_point)
                self.decision_boundary_points_2d.append(db_point2d)
                i += 1
                retries = 0

            edges, gap_distances, gap_probability_scores = (
                self._get_sorted_db_keypoint_distances()
            )  # reload gaps

        self.decision_boundary_points = np.array(self.decision_boundary_points)
        self.decision_boundary_points_2d = np.array(self.decision_boundary_points_2d)

        return self

    def plot(self):
        # decision boundary
        decision_boundary_x = self.decision_boundary_points_2d[:, 0]
        decision_boundary_y = self.decision_boundary_points_2d[:, 1]
        decision_boundary_color = "Decision Boundary"

        # training data
        train_x = self.X2d[self.train_idx, 0]
        train_y = self.X2d[self.train_idx, 1]
        train_color = ["Class 1 - Train Set"
                if self.y_pred[self.train_idx[i]] == self.y[self.train_idx[i]] == 1
                else (
                    "Class 2 - Train Set"
                    if self.y_pred[self.train_idx[i]] == self.y[self.train_idx[i]] == 0
                    else "Misclassified")
                for i in range(len(self.train_idx))]

        # testing data
        test_x = self.X2d[self.test_idx, 0]
        test_y = self.X2d[self.test_idx, 1]
        test_color = ["Class 1 - Test Set"
                if self.y_pred[self.test_idx[i]] == self.y[self.test_idx[i]] == 1
                else (
                    "Class 2 - Test Set"
                    if self.y_pred[self.test_idx[i]] == self.y[self.test_idx[i]] == 0
                    else "Misclassified")
                for i in range(len(self.test_idx))]
        return (
            decision_boundary_x,
            decision_boundary_y,
            decision_boundary_color,
            train_x,
            train_y,
            train_color,
            test_x,
            test_y,
            test_color
        )

    def decision_boundary_distance(self, x, grad=0):
        return np.abs(0.5 - self.classifier.predict_proba([x])[0][1])

    def get_decision_boundary_keypoints(self):
        if len(self.decision_boundary_points) == 0:
            raise Exception("Please call the fit method first!")
        return self.decision_boundary_points, self.decision_boundary_points_2d

    def _get_sorted_db_keypoint_distances(self, N=None):
        if N == None:
            N = self.n_interpolated_keypoints
        edges = minimum_spanning_tree(
            squareform(pdist(self.decision_boundary_points_2d))
        )
        edged = np.array(
            [
                euclidean(
                    self.decision_boundary_points_2d[u],
                    self.decision_boundary_points_2d[v],
                )
                for u, v in edges
            ]
        )
        gap_edge_idx = np.argsort(edged)[::-1][: int(N)]
        edges = edges[gap_edge_idx]
        gap_distances = np.square(edged[gap_edge_idx])
        gap_probability_scores = gap_distances / np.sum(gap_distances)
        return edges, gap_distances, gap_probability_scores

    def _linear_decision_boundary_optimization(
        self,
        from_idx,
        to_idx,
        all_combinations=True,
        retry_neighbor_if_failed=True,
        step=None,
        suppress_output=True,
    ):
        step_str = (
            ("Step " + str(step) + "/" + str(self.steps) + ":") if step != None else ""
        )

        retries = 4 if retry_neighbor_if_failed else 1
        for i in range(len(from_idx)):
            n = len(to_idx) if all_combinations else 1
            for j in range(n):
                from_i = from_idx[i]
                to_i = to_idx[j] if all_combinations else to_idx[i]
                for k in range(retries):
                    if k == 0:
                        from_point = self.Xminor[from_i]
                        to_point = self.Xmajor[to_i]
                    else:
                        # first attempt failed, try nearest neighbors of source and destination
                        # point instead
                        _, idx = self.nn_model_2d_minorityclass.kneighbors(
                            [self.Xminor2d[from_i]]
                        )
                        from_point = self.Xminor[idx[0][k // 2]]
                        _, idx = self.nn_model_2d_minorityclass.kneighbors(
                            [self.Xmajor2d[to_i]]
                        )
                        to_point = self.Xmajor[idx[0][k % 2]]

                    if euclidean(from_point, to_point) == 0:
                        break  # no decision boundary between equivalent points

                    db_point = self._find_decision_boundary_along_line(
                        from_point,
                        to_point,
                        penalize_tangent_distance=self.penalties_enabled,
                        penalize_extremes=self.penalties_enabled,
                    )

                    if (
                        self.decision_boundary_distance(db_point)
                        <= self.acceptance_threshold
                    ):
                        db_point2d = dimensionality_reduction.transform(
                            [db_point]
                        )[0]
                        if (
                            db_point2d[0] >= self.X2d_xmin
                            and db_point2d[0] <= self.X2d_xmax
                            and db_point2d[1] >= self.X2d_ymin
                            and db_point2d[1] <= self.X2d_ymax
                        ):
                            self.decision_boundary_points.append(db_point)
                            self.decision_boundary_points_2d.append(db_point2d)
                            break
                        else:
                            if self.verbose and not suppress_output:
                                msg = "{} {}/{}: Rejected decision boundary keypoint (outside of plot area)"

    def _find_decision_boundary_along_line(
        self,
        from_point,
        to_point,
        penalize_extremes=False,
        penalize_tangent_distance=False,
    ):
        def objective(l, grad=0):
            # interpolate between source and destionation; calculate distance from
            # decision boundary
            X = from_point + l[0] * (to_point - from_point)
            error = self.decision_boundary_distance(X)

            if penalize_tangent_distance:
                # distance from tangent between class1 and class0 point in 2d space
                x0, y0 = dimensionality_reduction.transform([X])[0]
                x1, y1 = dimensionality_reduction.transform([from_point])[0]
                x2, y2 = dimensionality_reduction.transform([to_point])[0]
                error += (
                    1e-12
                    * np.abs((y2 - y1) * x0 - (x2 - x1) * y0 + x2 * y1 - y2 * x1)
                    / np.sqrt((y2 - y1) ** 2 + (x2 - x1) ** 2)
                )

            if penalize_extremes:
                error += 1e-8 * np.abs(0.5 - l[0])

            return error

        optimizer = self._get_optimizer()
        optimizer.set_min_objective(objective)
        cl = optimizer.optimize([random.random()])
        db_point = from_point + cl[0] * (to_point - from_point)
        return db_point

    def _find_decision_boundary_on_hypersphere(self, centroid, R, penalize_known=False):
        def objective(phi, grad=0):
            # search on hypersphere surface in polar coordinates - map back to cartesian
            cx = centroid + polar_to_cartesian(phi, R)
            try:
                cx2d = dimensionality_reduction.transform([cx])[0]
                error = self.decision_boundary_distance(cx)
                if penalize_known:
                    # slight penalty for being too close to already known decision boundary
                    # keypoints
                    db_distances = [
                        euclidean(cx2d, self.decision_boundary_points_2d[k])
                        for k in range(len(self.decision_boundary_points_2d))
                    ]
                    error += (
                        1e-8
                        * (
                            (self.mean_2d_dist - np.min(db_distances))
                            / self.mean_2d_dist
                        )
                        ** 2
                    )
                return error
            except Exception as ex:
                wandb.termerror("Error in objective function:", ex)
                return np.infty

        optimizer = self._get_optimizer(
            D=self.X.shape[1] - 1,
            upper_bound=2 * np.pi,
            iteration_budget=self.hypersphere_iteration_budget,
        )
        optimizer.set_min_objective(objective)
        db_phi = optimizer.optimize(
            [random.random() * 2 * np.pi for k in range(self.X.shape[1] - 1)]
        )
        db_point = centroid + polar_to_cartesian(db_phi, R)
        return db_point

    def _get_optimizer(self, D=1, upper_bound=1, iteration_budget=None):
        if iteration_budget == None:
            iteration_budget = self.linear_iteration_budget

        '''
        opt = nlopt.opt(nlopt.GN_DIRECT_L_RAND, D)
        opt.set_stopval(self.acceptance_threshold/10.0)
        opt.set_ftol_rel(1e-5)
        opt.set_maxeval(iteration_budget)
        opt.set_lower_bounds(0)
        opt.set_upper_bounds(upper_bound)

        return opt
        '''

def minimum_spanning_tree(X, copy_X=True):
    if copy_X:
        X = X.copy()

    if X.shape[0] != X.shape[1]:
        raise ValueError("X needs to be square matrix of edge weights")
    n_vertices = X.shape[0]
    spanning_edges = []

    # initialize with node 0:
    visited_vertices = [0]
    num_visited = 1
    # exclude self connections:
    diag_indices = np.arange(n_vertices)
    X[diag_indices, diag_indices] = np.inf

    while num_visited != n_vertices:
        new_edge = np.argmin(X[visited_vertices], axis=None)
        # 2d encoding of new_edge from flat, get correct indices
        new_edge = divmod(new_edge, n_vertices)
        new_edge = [visited_vertices[new_edge[0]], new_edge[1]]
        # add edge to tree
        spanning_edges.append(new_edge)
        visited_vertices.append(new_edge[1])
        # remove all edges inside current tree
        X[visited_vertices, new_edge[1]] = np.inf
        X[new_edge[1], visited_vertices] = np.inf
        num_visited += 1
    return np.vstack(spanning_edges)


def polar_to_cartesian(arr, r):
    a = np.concatenate((np.array([2 * np.pi]), arr))
    si = np.sin(a)
    si[0] = 1
    si = np.cumprod(si)
    co = np.cos(a)
    co = np.roll(co, -1)
    return si * co * r



import pytest
from sklearn.naive_bayes import MultinomialNB
import wandb
from wandb.sklearn import learning_curve, roc, confusion_matrix, precision_recall

@pytest.fixture
def dummy_classifier(request):
    nb = MultinomialNB()
    x_train = [[1,2],[1,2],[1,2],[1,2],[2,3],[3,4],[3,4],[3,4],[3,4],[3,4],[3,4]]
    y_train = [0,0,0,0,0,1,1,1,1,1,1]
    nb.fit(x_train, y_train)
    x_test = [[4,5], [5,6]]
    y_test = [0,1]
    y_probas = nb.predict_proba(x_test)
    y_pred = nb.predict(x_test)
    return (nb, x_train, y_train, x_test, y_test, y_pred, y_probas)

def test_learning_curve(dummy_classifier):
    (nb, x_train, y_train, x_test, y_test, y_pred, y_probas) = dummy_classifier
    lc_table = learning_curve(nb, x_train, y_train)
    assert(len(lc_table.value.data) == 10)
    assert(lc_table.value.data[0][0] == 'train')
    assert(lc_table.value.data[1][0] == 'test')

def test_roc(dummy_classifier):
    (nb, x_train, y_train, x_test, y_test, y_pred, y_probas) = dummy_classifier
    lc_table = learning_curve(nb, x_train, y_train)
    r = roc(y_test, y_probas)

    assert(r.value.data[0] == [0, 0.0, 0.0])

def test_confusion_matrix(dummy_classifier):
    (nb, x_train, y_train, x_test, y_test, y_pred, y_probas) = dummy_classifier
    cm = confusion_matrix(y_test, y_pred)
    assert(len(cm.value.data)==4)
    assert(cm.value.data[0]== [0,0,0])

def test_precision_recall(dummy_classifier):
    (nb, x_train, y_train, x_test, y_test, y_pred, y_probas) = dummy_classifier
    pr = precision_recall(y_test, y_probas)

    assert(pr.value.data[0]== [0, 1.0, 1.0])



import pytest
from .utils import git_repo
import os
import glob
import sys
import six
from click.testing import CliRunner
import wandb
import types
import subprocess
from wandb import env
from wandb import util
from wandb.meta import Meta
from wandb.apis import InternalApi


def test_meta(git_repo, mocker):
    mocker.patch.object(sys, 'argv', ["foo", "bar"])
    meta = Meta(InternalApi())
    meta.write()
    print(meta.data)
    assert meta.data["cpu_count"] > 0
    assert meta.data["git"]["commit"]
    assert meta.data["heartbeatAt"]
    assert meta.data["startedAt"]
    assert meta.data["host"]
    assert meta.data["root"] == os.getcwd()
    assert meta.data["python"]
    assert meta.data["program"]
    assert meta.data["executable"]
    assert meta.data["args"] == ["bar"]
    assert meta.data["state"] == "running"
    assert meta.data["username"]
    assert meta.data["os"]


def test_anonymous_redaction(mocker):
    mocker.patch.object(sys, 'argv', ["foo", "bar"])
    api = InternalApi()
    api.set_setting('anonymous', 'true')

    meta = Meta(api, "wandb")
    meta.write()

    print(meta.data)
    assert "host" not in meta.data
    assert "username" not in meta.data
    assert "executable" not in meta.data
    assert "email" not in meta.data
    assert "root" not in meta.data


def test_disable_code(git_repo):
    os.environ[env.DISABLE_CODE] = "true"
    meta = Meta(InternalApi())
    assert meta.data.get("git") is None
    del os.environ[env.DISABLE_CODE]


def test_colab(mocker, monkeypatch):
    with CliRunner().isolated_filesystem():
        mocker.patch('wandb._get_python_type', lambda: "jupyter")
        with open("test.ipynb", "w") as f:
            f.write("{}")
        module = types.ModuleType("fake_jupyter")
        module.notebook_metadata = lambda: {"path": "fileId=123abc", "name": "test.ipynb", "root": os.getcwd()}
        monkeypatch.setattr(wandb, 'jupyter', module)
        meta = Meta(InternalApi())
        assert meta.data["colab"] == "https://colab.research.google.com/drive/123abc"
        assert meta.data["program"] == "test.ipynb"
        assert meta.data["codeSaved"]
        assert os.path.exists("code/test.ipynb")


def test_git_untracked_notebook_env(monkeypatch, git_repo, mocker):
    mocker.patch('wandb._get_python_type', lambda: "jupyter")
    with open("test.ipynb", "w") as f:
        f.write("{}")
    os.environ[env.NOTEBOOK_NAME] = "test.ipynb"
    meta = Meta(InternalApi())
    assert meta.data["program"] == "test.ipynb"
    assert meta.data["codeSaved"]
    assert os.path.exists("code/test.ipynb")
    os.environ[env.NOTEBOOK_NAME]


def test_git_untracked_notebook_env_subdir(monkeypatch, git_repo, mocker):
    mocker.patch('wandb._get_python_type', lambda: "jupyter")
    util.mkdir_exists_ok("sub")
    with open("sub/test.ipynb", "w") as f:
        f.write("{}")
    os.environ[env.NOTEBOOK_NAME] = "sub/test.ipynb"
    meta = Meta(InternalApi())
    assert meta.data["program"] == "sub/test.ipynb"
    assert meta.data["codeSaved"]
    assert os.path.exists("code/sub/test.ipynb")
    os.environ[env.NOTEBOOK_NAME]


def test_git_tracked_notebook_env(monkeypatch, git_repo, mocker):
    mocker.patch('wandb._get_python_type', lambda: "jupyter")
    with open("test.ipynb", "w") as f:
        f.write("{}")
    subprocess.check_call(['git', 'add', 'test.ipynb'])
    os.environ[env.NOTEBOOK_NAME] = "test.ipynb"
    meta = Meta(InternalApi())
    assert meta.data["program"] == "test.ipynb"
    assert not meta.data.get("codeSaved")
    assert not os.path.exists("code/test.ipynb")
    os.environ[env.NOTEBOOK_NAME]


def test_meta_cuda(mocker):
    mocker.patch('wandb.meta.os.path.exists', lambda path: True)

    def magic(path, mode="w"):
        if "cuda/version.txt" in path:
            return six.StringIO("CUDA Version 9.0.176")
        else:
            return open(path, mode=mode)
    mocker.patch('wandb.meta.open', magic)
    meta = Meta(InternalApi())
    assert meta.data["cuda"] == "9.0.176"


def test_meta_thread(git_repo):
    meta = Meta(InternalApi(), "wandb")
    meta.start()
    meta.shutdown()
    print("GO", glob.glob("**"))
    assert os.path.exists("wandb/wandb-metadata.json")



import pytest
import numpy as np
import lightgbm as lgb
import wandb
import sys
from wandb import wandb_run
from wandb.lightgbm import wandb_callback

# Tests which rely on row history in memory should set `History.keep_rows = True`
from wandb.history import History
History.keep_rows = True


@pytest.fixture
def dummy_data(request):
    data = np.array([[1, 2], [3, 4]])
    label = np.array([0, 1])
    dtrain = lgb.Dataset(data, label=label)
    return dtrain

@pytest.mark.skipif(sys.version_info < (3, 6), reason="lightgbm was segfaulting in CI")
def test_basic_lightgbm(dummy_data, wandb_init_run):
    param = {'max_depth': 2, 'eta': 1}
    num_round = 2
    lgb.train(param, dummy_data, num_round, callbacks=[wandb_callback()])



import pytest
import datetime
import requests

from wandb import retry
from wandb import util
from wandb.apis import CommError


class FailException(Exception):
    pass


def fail_for_n_function(n):
    call_num = [0]  # use a list so we can modify in this scope

    def fn(an_arg):
        print(an_arg)
        try:
            if call_num[0] < n:
                raise retry.TransientException(
                    exc=FailException('Failed at call_num: %s' % call_num))
        finally:
            call_num[0] += 1
        return True
    return fn


def test_fail_for_n_function():
    failing_fn = fail_for_n_function(3)
    with pytest.raises(retry.TransientException):
        failing_fn('hello')
    with pytest.raises(retry.TransientException):
        failing_fn('hello')
    with pytest.raises(retry.TransientException):
        failing_fn('hello')
    assert failing_fn('hello')


def test_retry_with_success():
    failing_fn = fail_for_n_function(3)
    fn = retry.Retry(failing_fn)
    fn('hello', retry_timedelta=datetime.timedelta(days=1), retry_sleep_base=0.001)
    assert fn.num_iters == 3


def test_retry_with_timeout():
    failing_fn = fail_for_n_function(10000)
    fn = retry.Retry(failing_fn)
    with pytest.raises(retry.TransientException):
        fn('hello', retry_timedelta=datetime.timedelta(
            0, 0, 0, 50), retry_sleep_base=0.001)


def test_retry_with_noauth_401(capsys):
    def fail():
        res = requests.Response()
        res.status_code = 401
        raise retry.TransientException(exc=requests.HTTPError(response=res))
    fn = retry.Retry(fail, check_retry_fn=util.no_retry_auth)
    with pytest.raises(CommError) as excinfo:
        fn()
    assert excinfo.value.message == 'Invalid or missing api_key.  Run wandb login'


def test_retry_with_noauth_403(capsys):
    def fail():
        res = requests.Response()
        res.status_code = 403
        raise retry.TransientException(exc=requests.HTTPError(response=res))
    fn = retry.Retry(fail, check_retry_fn=util.no_retry_auth)
    with pytest.raises(CommError) as excinfo:
        fn()
    assert excinfo.value.message == 'Permission denied, ask the project owner to grant you access'



# -*- coding: utf-8 -*-
from __future__ import print_function

import click
import copy
from functools import wraps
import glob
import io
import json
import logging
import netrc
import os
import random
import re
import requests
import shlex
import signal
import socket
import stat
import subprocess
import sys
import textwrap
import time
import traceback
import yaml
import threading
import random
import platform
import datetime
import shutil
# pycreds has a find_executable that works in windows
from dockerpycreds.utils import find_executable

from wandb import util

from click.utils import LazyFile
from click.exceptions import BadParameter, ClickException, Abort
# whaaaaat depends on prompt_toolkit < 2, ipython now uses > 2 so we vendored for now
# DANGER this changes the sys.path so we should never do this in a user script
whaaaaat = util.vendor_import("whaaaaat")
import six
from six.moves import BaseHTTPServer, urllib, configparser
import socket

from .core import termlog

import wandb
from wandb.apis import InternalApi
from wandb.wandb_config import Config
from wandb import wandb_agent
from wandb import wandb_controller
from wandb import env
from wandb import wandb_run
from wandb import wandb_dir
from wandb import run_manager
from wandb import Error
from wandb.magic_impl import magic_install

DOCS_URL = 'http://docs.wandb.com/'
logger = logging.getLogger(__name__)


class ClickWandbException(ClickException):
    def format_message(self):
        log_file = util.get_log_file_path()
        orig_type = '{}.{}'.format(self.orig_type.__module__,
                                   self.orig_type.__name__)
        if issubclass(self.orig_type, Error):
            return click.style(str(self.message), fg="red")
        else:
            return ('An Exception was raised, see %s for full traceback.\n'
                    '%s: %s' % (log_file, orig_type, self.message))


class CallbackHandler(BaseHTTPServer.BaseHTTPRequestHandler):
    """Simple callback handler that stores query string parameters and
    shuts down the server.
    """

    def do_GET(self):
        self.server.result = urllib.parse.parse_qs(
            self.path.split("?")[-1])
        self.send_response(200)
        self.end_headers()
        self.wfile.write(b'Success')
        self.server.stop()

    def log_message(self, format, *args):
        pass


class LocalServer():
    """A local HTTP server that finds an open port and listens for a callback.
    The urlencoded callback url is accessed via `.qs` the query parameters passed
    to the callback are accessed via `.result`
    """

    def __init__(self):
        self.blocking = True
        self.port = 8666
        self.connect()
        self._server.result = {}
        self._server.stop = self.stop

    def connect(self, attempts=1):
        try:
            self._server = BaseHTTPServer.HTTPServer(
                ('127.0.0.1', self.port), CallbackHandler)
        except socket.error:
            if attempts < 5:
                self.port += random.randint(1, 1000)
                self.connect(attempts + 1)
            else:
                logging.info(
                    "Unable to start local server, proceeding manually")

                class FakeServer():
                    def serve_forever(self):
                        pass
                self._server = FakeServer()

    def qs(self):
        return urllib.parse.urlencode({
            "callback": "http://127.0.0.1:{}/callback".format(self.port)})

    @property
    def result(self):
        return self._server.result

    def start(self, blocking=True):
        self.blocking = blocking
        if self.blocking:
            self._server.serve_forever()
        else:
            t = threading.Thread(target=self._server.serve_forever)
            t.daemon = True
            t.start()

    def stop(self, *args):
        t = threading.Thread(target=self._server.shutdown)
        t.daemon = True
        t.start()
        if not self.blocking:
            os.kill(os.getpid(), signal.SIGINT)


def display_error(func):
    """Function decorator for catching common errors and re-raising as wandb.Error"""
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except wandb.Error as e:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            lines = traceback.format_exception(
                exc_type, exc_value, exc_traceback)
            logger.error(''.join(lines))
            click_exc = ClickWandbException(e)
            click_exc.orig_type = exc_type
            six.reraise(ClickWandbException, click_exc, sys.exc_info()[2])
    return wrapper


def prompt_for_project(ctx, entity):
    """Ask the user for a project, creating one if necessary."""
    result = ctx.invoke(projects, entity=entity, display=False)

    try:
        if len(result) == 0:
            project = click.prompt("Enter a name for your first project")
            #description = editor()
            project = api.upsert_project(project, entity=entity)["name"]
        else:
            project_names = [project["name"] for project in result]
            question = {
                'type': 'list',
                'name': 'project_name',
                'message': "Which project should we use?",
                'choices': project_names + ["Create New"]
            }
            result = whaaaaat.prompt([question])
            if result:
                project = result['project_name']
            else:
                project = "Create New"
            # TODO: check with the server if the project exists
            if project == "Create New":
                project = click.prompt(
                    "Enter a name for your new project", value_proc=api.format_project)
                #description = editor()
                project = api.upsert_project(project, entity=entity)["name"]

    except wandb.apis.CommError as e:
        raise ClickException(str(e))

    return project


def editor(content='', marker='# Enter a description, markdown is allowed!\n'):
    message = click.edit(content + '\n\n' + marker)
    if message is not None:
        return message.split(marker, 1)[0].rstrip('\n')


api = InternalApi()


# Some commands take project/entity etc. as arguments. We provide default
# values for those arguments from the current project configuration, as
# returned by api.settings()
CONTEXT = dict(default_map=api.settings())


class RunGroup(click.Group):
    @display_error
    def get_command(self, ctx, cmd_name):
        # TODO: check if cmd_name is a file in the current dir and not require `run`?
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv

        return None


@click.command(cls=RunGroup, invoke_without_command=True)
@click.version_option(version=wandb.__version__)
@click.pass_context
def cli(ctx):
    """Weights & Biases.

    Run "wandb docs" for full documentation.
    """
    wandb.try_to_set_up_global_logging()
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())


@cli.command(context_settings=CONTEXT, help="List projects")
@click.option("--entity", "-e", default=None, envvar=env.ENTITY, help="The entity to scope the listing to.")
@display_error
def projects(entity, display=True):
    projects = api.list_projects(entity=entity)
    if len(projects) == 0:
        message = "No projects found for %s" % entity
    else:
        message = 'Latest projects for "%s"' % entity
    if display:
        click.echo(click.style(message, bold=True))
        for project in projects:
            click.echo("".join(
                (click.style(project['name'], fg="blue", bold=True),
                 " - ",
                 str(project['description'] or "").split("\n")[0])
            ))
    return projects


@cli.command(context_settings=CONTEXT, help="List runs in a project")
@click.pass_context
@click.option("--project", "-p", default=None, envvar=env.PROJECT, help="The project you wish to list runs from.")
@click.option("--entity", "-e", default=None, envvar=env.ENTITY, help="The entity to scope the listing to.")
@display_error
def runs(ctx, project, entity):
    click.echo(click.style('Latest runs for project "%s"' %
                           project, bold=True))
    if project is None:
        project = prompt_for_project(ctx, project)
    runs = api.list_runs(project, entity=entity)
    for run in runs:
        click.echo("".join(
            (click.style(run['name'], fg="blue", bold=True),
             " - ",
             (run['description'] or "").split("\n")[0])
        ))


@cli.command(context_settings=CONTEXT, help="List local & remote file status")
@click.argument("run", envvar=env.RUN_ID)
@click.option("--settings/--no-settings", help="Show the current settings", default=True)
@display_error
def status(run, settings):
    logged_in = bool(api.api_key)
    if not os.path.isdir(wandb_dir()):
        if logged_in:
            msg = "Directory not initialized. Please run %s to get started." % click.style(
                "wandb init", bold=True)
        else:
            msg = "You are not logged in. Please run %s to get started." % click.style(
                "wandb login", bold=True)
        termlog(msg)
    elif settings:
        click.echo(click.style("Logged in?", bold=True) + " %s" % logged_in)
        click.echo(click.style("Current Settings", bold=True))
        settings = api.settings()
        click.echo(json.dumps(
            settings,
            sort_keys=True,
            indent=2,
            separators=(',', ': ')
        ))


@cli.command(context_settings=CONTEXT, help="Restore code, config and docker state for a run")
@click.pass_context
@click.argument("run", envvar=env.RUN_ID)
@click.option("--no-git", is_flag=True, default=False, help="Skupp")
@click.option("--branch/--no-branch", default=True, help="Whether to create a branch or checkout detached")
@click.option("--project", "-p", envvar=env.PROJECT, help="The project you wish to upload to.")
@click.option("--entity", "-e", envvar=env.ENTITY, help="The entity to scope the listing to.")
@display_error
def restore(ctx, run, no_git, branch, project, entity):
    if ":" in run:
        if "/" in run:
            entity, rest = run.split("/", 1)
        else:
            rest = run
        project, run = rest.split(":", 1)
    elif run.count("/") > 1:
        entity, run = run.split("/", 1)

    project, run = api.parse_slug(run, project=project)
    commit, json_config, patch_content, metadata = api.run_config(
        project, run=run, entity=entity)
    repo = metadata.get("git", {}).get("repo")
    image = metadata.get("docker")
    RESTORE_MESSAGE = """`wandb restore` needs to be run from the same git repository as the original run.
Run `git clone %s` and restore from there or pass the --no-git flag.""" % repo
    if no_git:
        commit = None
    elif not api.git.enabled:
        if repo:
            raise ClickException(RESTORE_MESSAGE)
        elif image:
            wandb.termlog("Original run has no git history.  Just restoring config and docker")

    if commit and api.git.enabled:
        subprocess.check_call(['git', 'fetch', '--all'])
        try:
            api.git.repo.commit(commit)
        except ValueError:
            wandb.termlog("Couldn't find original commit: {}".format(commit))
            commit = None
            files = api.download_urls(project, run=run, entity=entity)
            for filename in files:
                if filename.startswith('upstream_diff_') and filename.endswith('.patch'):
                    commit = filename[len('upstream_diff_'):-len('.patch')]
                    try:
                        api.git.repo.commit(commit)
                    except ValueError:
                        commit = None
                    else:
                        break

            if commit:
                wandb.termlog(
                    "Falling back to upstream commit: {}".format(commit))
                patch_path, _ = api.download_write_file(files[filename])
            else:
                raise ClickException(RESTORE_MESSAGE)
        else:
            if patch_content:
                patch_path = os.path.join(wandb.wandb_dir(), 'diff.patch')
                with open(patch_path, "w") as f:
                    f.write(patch_content)
            else:
                patch_path = None

        branch_name = "wandb/%s" % run
        if branch and branch_name not in api.git.repo.branches:
            api.git.repo.git.checkout(commit, b=branch_name)
            wandb.termlog("Created branch %s" %
                       click.style(branch_name, bold=True))
        elif branch:
            wandb.termlog(
                "Using existing branch, run `git branch -D %s` from master for a clean checkout" % branch_name)
            api.git.repo.git.checkout(branch_name)
        else:
            wandb.termlog("Checking out %s in detached mode" % commit)
            api.git.repo.git.checkout(commit)

        if patch_path:
            # we apply the patch from the repository root so git doesn't exclude
            # things outside the current directory
            root = api.git.root
            patch_rel_path = os.path.relpath(patch_path, start=root)
            # --reject is necessary or else this fails any time a binary file
            # occurs in the diff
            # we use .call() instead of .check_call() for the same reason
            # TODO(adrian): this means there is no error checking here
            subprocess.call(['git', 'apply', '--reject',
                             patch_rel_path], cwd=root)
            wandb.termlog("Applied patch")

    # TODO: we should likely respect WANDB_DIR here.
    util.mkdir_exists_ok("wandb")
    config = Config(run_dir="wandb")
    config.load_json(json_config)
    config.persist()
    wandb.termlog("Restored config variables to %s" % config._config_path())
    if image:
        if not metadata["program"].startswith("<") and metadata.get("args") is not None:
            # TODO: we may not want to default to python here.
            runner = util.find_runner(metadata["program"]) or ["python"]
            command = runner + [metadata["program"]] + metadata["args"]
            cmd = " ".join(command)
        else:
            wandb.termlog("Couldn't find original command, just restoring environment")
            cmd = None
        wandb.termlog("Docker image found, attempting to start")
        ctx.invoke(docker, docker_run_args=[image], cmd=cmd)

    return commit, json_config, patch_content, repo, metadata


@cli.command(context_settings=CONTEXT, help="Upload an offline training directory to W&B")
@click.pass_context
@click.argument("path", nargs=-1, type=click.Path(exists=True))
@click.option("--id", envvar=env.RUN_ID, help="The run you want to upload to.")
@click.option("--project", "-p", envvar=env.PROJECT, help="The project you want to upload to.")
@click.option("--entity", "-e", envvar=env.ENTITY, help="The entity to scope to.")
@click.option("--ignore", help="A comma seperated list of globs to ignore syncing with wandb.")
@display_error
def sync(ctx, path, id, project, entity, ignore):
    if api.api_key is None:
        ctx.invoke(login)

    if ignore:
        globs = ignore.split(",")
    else:
        globs = None

    path = path[0] if len(path) > 0 else os.getcwd()
    if os.path.isfile(path):
        raise ClickException("path must be a directory")
    wandb_dir = os.path.join(path, "wandb")
    run_paths = glob.glob(os.path.join(wandb_dir, "*run-*"))
    if len(run_paths) == 0:
        run_paths = glob.glob(os.path.join(path, "*run-*"))
    if len(run_paths) > 0:
        for run_path in run_paths:
            wandb_run.Run.from_directory(run_path,
                                         run_id=run_path.split("-")[-1], project=project, entity=entity, ignore_globs=globs)
    else:
        wandb_run.Run.from_directory(
            path, run_id=id, project=project, entity=entity, ignore_globs=globs)


@cli.command(context_settings=CONTEXT, help="Pull files from Weights & Biases")
@click.argument("run", envvar=env.RUN_ID)
@click.option("--project", "-p", envvar=env.PROJECT, help="The project you want to download.")
@click.option("--entity", "-e", default="models", envvar=env.ENTITY, help="The entity to scope the listing to.")
@display_error
def pull(run, project, entity):
    project, run = api.parse_slug(run, project=project)

    urls = api.download_urls(project, run=run, entity=entity)
    if len(urls) == 0:
        raise ClickException("Run has no files")
    click.echo("Downloading: {project}/{run}".format(
        project=click.style(project, bold=True), run=run
    ))

    for name in urls:
        if api.file_current(name, urls[name]['md5']):
            click.echo("File %s is up to date" % name)
        else:
            length, response = api.download_file(urls[name]['url'])
            # TODO: I had to add this because some versions in CI broke click.progressbar
            sys.stdout.write("File %s\r" % name)
            dirname = os.path.dirname(name)
            if dirname != '':
                wandb.util.mkdir_exists_ok(dirname)
            with click.progressbar(length=length, label='File %s' % name,
                                   fill_char=click.style('&', fg='green')) as bar:
                with open(name, "wb") as f:
                    for data in response.iter_content(chunk_size=4096):
                        f.write(data)
                        bar.update(len(data))


@cli.command(context_settings=CONTEXT, help="Login to Weights & Biases")
@click.argument("key", nargs=-1)
@click.option("--host", default=None, help="Login to a specific instance of W&B")
@click.option("--browser/--no-browser", default=True, help="Attempt to launch a browser for login")
@click.option("--anonymously", is_flag=True, help="Log in anonymously")
@display_error
def login(key, host, anonymously, server=LocalServer(), browser=True, no_offline=False):
    global api
    if host == "https://api.wandb.ai":
        api.clear_setting("base_url", globally=True)
    elif host:
        if not host.startswith("http"):
            raise ClickException("host must start with http(s)://")
        api.set_setting("base_url", host, globally=True)

    key = key[0] if len(key) > 0 else None

    # Import in here for performance reasons
    import webbrowser
    browser = util.launch_browser(browser)

    def get_api_key_from_browser(signup=False):
        if not browser:
            return None
        query = '?signup=true' if signup else ''
        webbrowser.open_new_tab('{}/authorize{}'.format(api.app_url, query))
        #Getting rid of the server for now.  We would need to catch Abort from server.stop and deal accordingly
        #server.start(blocking=False)
        #if server.result.get("key"):
        #    return server.result["key"][0]
        return None

    if key:
        util.set_api_key(api, key)
    else:
        if anonymously:
            os.environ[env.ANONYMOUS] = "must"
        # Don't allow signups or dryrun for local
        local = host != None or host != "https://api.wandb.ai"
        key = util.prompt_api_key(api, input_callback=click.prompt,
            browser_callback=get_api_key_from_browser, no_offline=no_offline, local=local)

    if key:
        api.clear_setting('disabled')
        click.secho("Successfully logged in to Weights & Biases!", fg="green")
    elif not no_offline:
        api.set_setting('disabled', 'true')
        click.echo("Disabling Weights & Biases. Run 'wandb login' again to re-enable.")

    # reinitialize API to create the new client
    api = InternalApi()

    return key


@cli.command(context_settings=CONTEXT, help="Configure a directory with Weights & Biases")
@click.pass_context
@display_error
def init(ctx):
    from wandb import _set_stage_dir, __stage_dir__, wandb_dir
    if __stage_dir__ is None:
        _set_stage_dir('wandb')
    if os.path.isdir(wandb_dir()) and os.path.exists(os.path.join(wandb_dir(), "settings")):
        click.confirm(click.style(
            "This directory has been configured previously, should we re-configure it?", bold=True), abort=True)
    else:
        click.echo(click.style(
            "Let's setup this directory for W&B!", fg="green", bold=True))

    if api.api_key is None:
        ctx.invoke(login)

    viewer = api.viewer()

    # Viewer can be `None` in case your API information became invalid, or
    # in testing if you switch hosts.
    if not viewer:
        click.echo(click.style(
            "Your login information seems to be invalid: can you log in again please?", fg="red", bold=True))
        ctx.invoke(login)

    # This shouldn't happen.
    viewer = api.viewer()
    if not viewer:
        click.echo(click.style(
            "We're sorry, there was a problem logging you in. Please send us a note at support@wandb.com and tell us how this happened.", fg="red", bold=True))
        sys.exit(1)

    # At this point we should be logged in successfully.
    if len(viewer["teams"]["edges"]) > 1:
        team_names = [e["node"]["name"] for e in viewer["teams"]["edges"]]
        question = {
            'type': 'list',
            'name': 'team_name',
            'message': "Which team should we use?",
            'choices': team_names + ["Manual Entry"]
        }
        result = whaaaaat.prompt([question])
        # result can be empty on click
        if result:
            entity = result['team_name']
        else:
            entity = "Manual Entry"
        if entity == "Manual Entry":
            entity = click.prompt("Enter the name of the team you want to use")
    else:
        entity = click.prompt("What username or team should we use?",
                              default=viewer.get('entity', 'models'))

    # TODO: this error handling sucks and the output isn't pretty
    try:
        project = prompt_for_project(ctx, entity)
    except wandb.cli.ClickWandbException:
        raise ClickException('Could not find team: %s' % entity)

    api.set_setting('entity', entity)
    api.set_setting('project', project)
    api.set_setting('base_url', api.settings().get('base_url'))

    util.mkdir_exists_ok(wandb_dir())
    with open(os.path.join(wandb_dir(), '.gitignore'), "w") as file:
        file.write("*\n!settings")

    click.echo(click.style("This directory is configured!  Next, track a run:\n", fg="green") +
               textwrap.dedent("""\
        * In your training script:
            {code1}
            {code2}
        * then `{run}`.
        """).format(
        code1=click.style("import wandb", bold=True),
        code2=click.style("wandb.init(project=\"%s\")" % project, bold=True),
        run=click.style("python <train.py>", bold=True),
        # saving this here so I can easily put it back when we re-enable
        # push/pull
        # """
        # * Run `{push}` to manually add a file.
        # * Pull popular models into your project with: `{pull}`.
        # """
        # push=click.style("wandb push run_id weights.h5", bold=True),
        # pull=click.style("wandb pull models/inception-v4", bold=True)
    ))


@cli.command(context_settings=CONTEXT, help="Open documentation in a browser")
@click.pass_context
@display_error
def docs(ctx):
    import webbrowser
    if util.launch_browser():
        launched = webbrowser.open_new_tab(DOCS_URL)
    else:
        launched = False
    if launched:
        click.echo(click.style(
            "Opening %s in your default browser" % DOCS_URL, fg="green"))
    else:
        click.echo(click.style(
            "You can find our documentation here: %s" % DOCS_URL, fg="green"))


@cli.command("on", help="Ensure W&B is enabled in this directory")
@display_error
def on():
    wandb.ensure_configured()
    api = InternalApi()
    try:
        api.clear_setting('disabled')
    except configparser.Error:
        pass
    click.echo(
        "W&B enabled, running your script from this directory will now sync to the cloud.")


@cli.command("off", help="Disable W&B in this directory, useful for testing")
@display_error
def off():
    wandb.ensure_configured()
    api = InternalApi()
    try:
        api.set_setting('disabled', 'true')
        click.echo(
            "W&B disabled, running your script from this directory will only write metadata locally.")
    except configparser.Error as e:
        click.echo(
            'Unable to write config, copy and paste the following in your terminal to turn off W&B:\nexport WANDB_MODE=dryrun')


RUN_CONTEXT = copy.copy(CONTEXT)
RUN_CONTEXT['allow_extra_args'] = True
RUN_CONTEXT['ignore_unknown_options'] = True


@cli.command(context_settings=RUN_CONTEXT, help="Launch a job")
@click.pass_context
@click.argument('program')
@click.argument('args', nargs=-1)
@click.option('--id', default=None,
              help='Run id to use, default is to generate.')
@click.option('--resume', default='never', type=click.Choice(['never', 'must', 'allow']),
              help='Resume strategy, default is never')
@click.option('--dir', default=None,
              help='Files in this directory will be saved to wandb, defaults to wandb')
@click.option('--configs', default=None,
              help='Config file paths to load')
@click.option('--message', '-m', default=None, hidden=True,
              help='Message to associate with the run.')
@click.option('--name', default=None,
              help='Name of the run, default is auto generated.')
@click.option('--notes', default=None,
              help='Notes to associate with the run.')
@click.option("--show/--no-show", default=False,
              help="Open the run page in your default browser.")
@click.option('--tags', default=None,
              help='Tags to associate with the run (comma seperated).')
@click.option('--run_group', default=None,
              help='Run group to associate with the run.')
@click.option('--job_type', default=None,
              help='Job type to associate with the run.')
@display_error
def run(ctx, program, args, id, resume, dir, configs, message, name, notes, show, tags, run_group, job_type):
    wandb.ensure_configured()
    if configs:
        config_paths = configs.split(',')
    else:
        config_paths = []
    config = Config(config_paths=config_paths,
                    wandb_dir=dir or wandb.wandb_dir())
    tags = [tag for tag in tags.split(",") if tag] if tags else None

    # populate run parameters from env if not specified
    id = id or os.environ.get(env.RUN_ID)
    message = message or os.environ.get(env.DESCRIPTION)
    tags = tags or env.get_tags()
    run_group = run_group or os.environ.get(env.RUN_GROUP)
    job_type = job_type or os.environ.get(env.JOB_TYPE)
    name = name or os.environ.get(env.NAME)
    notes = notes or os.environ.get(env.NOTES)
    resume = resume or os.environ.get(env.RESUME)

    run = wandb_run.Run(run_id=id, mode='clirun',
                        config=config, description=message,
                        program=program, tags=tags,
                        group=run_group, job_type=job_type,
                        name=name, notes=notes,
                        resume=resume)
    run.enable_logging()

    environ = dict(os.environ)
    if configs:
        environ[env.CONFIG_PATHS] = configs
    if show:
        environ[env.SHOW_RUN] = 'True'

    if not run.api.api_key:
        util.prompt_api_key(run.api, input_callback=click.prompt)

    try:
        rm = run_manager.RunManager(run)
        rm.init_run(environ)
    except run_manager.Error:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        wandb.termerror('An Exception was raised during setup, see %s for full traceback.' %
                        util.get_log_file_path())
        wandb.termerror(str(exc_value))
        if 'permission' in str(exc_value):
            wandb.termerror(
                'Are you sure you provided the correct API key to "wandb login"?')
        lines = traceback.format_exception(
            exc_type, exc_value, exc_traceback)
        logger.error('\n'.join(lines))
        sys.exit(1)
    rm.run_user_process(program, args, environ)

@cli.command(context_settings=RUN_CONTEXT, help="Launch local W&B container (Experimental)")
@click.pass_context
@click.option('--port', '-p', default="8080", help="The host port to bind W&B local on")
@click.option('--daemon/--no-daemon', default=True, help="Run or don't run in daemon mode")
@click.option('--upgrade', is_flag=True, default=False, help="Upgrade to the most recent version")
@click.option('--edge', is_flag=True, default=False, help="Run the bleading edge", hidden=True)
@display_error
def local(ctx, port, daemon, upgrade, edge):
    if not find_executable('docker'):
        raise ClickException(
            "Docker not installed, install it from https://docker.com" )
    if wandb.docker.image_id("wandb/local") != wandb.docker.image_id_from_registry("wandb/local"):
        if upgrade:
            subprocess.call(["docker", "pull", "wandb/local"])
        else:
            wandb.termlog("A new version of W&B local is available, upgrade by calling `wandb local --upgrade`")
    running = subprocess.check_output(["docker", "ps", "--filter", "name=wandb-local", "--format", "{{.ID}}"])
    if running != b"":
        if upgrade:
            subprocess.call(["docker", "restart", "wandb-local"])
            exit(0)
        else:
            wandb.termerror("A container named wandb-local is already running, run `docker kill wandb-local` if you want to start a new instance")
            exit(1)
    image = "docker.pkg.github.com/wandb/core/local" if edge else "wandb/local"
    command = ['docker', 'run', '--rm', '-v', 'wandb:/vol', '-p', port+':8080', '--name', 'wandb-local']
    host = "http://localhost:%s" % port
    api.set_setting("base_url", host, globally=True)
    if daemon:
        command += ["-d"]
    command += [image]

    # DEVNULL is only in py3
    try:
        from subprocess import DEVNULL
    except ImportError:
        DEVNULL = open(os.devnull, 'wb')
    code = subprocess.call(command, stdout=DEVNULL)
    if daemon:
        if code != 0:
            wandb.termerror("Failed to launch the W&B local container, see the above error.")
            exit(1)
        else:
            wandb.termlog("W&B local started at http://localhost:%s \U0001F680" % port)
            wandb.termlog("You can stop the server by running `docker kill wandb-local`")
            if not api.api_key:
                ctx.invoke(login, host=host)

@cli.command(context_settings=RUN_CONTEXT)
@click.pass_context
@click.option('--keep', '-N', default=24, help="Keep runs created in the last N hours", type=int)
def gc(ctx, keep):
    """Garbage collector, cleans up your local run directory"""
    directory = wandb.wandb_dir()
    if not os.path.exists(directory):
        raise ClickException('No wandb directory found at %s' % directory)
    paths = glob.glob(directory+"/*run*")
    dates = [datetime.datetime.strptime(p.split("-")[1],'%Y%m%d_%H%M%S') for p in paths]
    since = datetime.datetime.utcnow() - datetime.timedelta(hours=keep)
    bad_paths = [paths[i] for i, d, in enumerate(dates) if d < since]
    if len(bad_paths) > 0:
        click.echo("Found {} runs, {} are older than {} hours".format(len(paths), len(bad_paths), keep))
        click.confirm(click.style(
                "Are you sure you want to remove %i runs?" % len(bad_paths), bold=True), abort=True)
        for path in bad_paths:
            shutil.rmtree(path)
        click.echo(click.style("Success!", fg="green"))
    else:
        click.echo(click.style("No runs older than %i hours found" % keep, fg="red"))

@cli.command(context_settings=RUN_CONTEXT, name="docker-run")
@click.pass_context
@click.argument('docker_run_args', nargs=-1)
@click.option('--help')
def docker_run(ctx, docker_run_args, help):
    """Simple docker wrapper that adds WANDB_API_KEY and WANDB_DOCKER to any docker run command.
    This will also set the runtime to nvidia if the nvidia-docker executable is present on the system
    and --runtime wasn't set.
    """
    args = list(docker_run_args)
    if len(args) > 0 and args[0] == "run":
        args.pop(0)
    if help or len(args) == 0:
        wandb.termlog("This commands adds wandb env variables to your docker run calls")
        subprocess.call(['docker', 'run'] + args + ['--help'])
        exit()
    #TODO: is this what we want?
    if len([a for a in args if a.startswith("--runtime")]) == 0 and find_executable('nvidia-docker'):
        args = ["--runtime", "nvidia"] + args
    #TODO: image_from_docker_args uses heuristics to find the docker image arg, there are likely cases
    #where this won't work
    image = util.image_from_docker_args(args)
    resolved_image = None
    if image:
        resolved_image = wandb.docker.image_id(image)
    if resolved_image:
        args = ['-e', 'WANDB_DOCKER=%s' % resolved_image] + args
    else:
        wandb.termlog("Couldn't detect image argument, running command without the WANDB_DOCKER env variable")
    if api.api_key:
        args = ['-e', 'WANDB_API_KEY=%s' % api.api_key] + args
    else:
        wandb.termlog("Not logged in, run `wandb login` from the host machine to enable result logging")
    subprocess.call(['docker', 'run'] + args)


@cli.command(context_settings=RUN_CONTEXT)
@click.pass_context
@click.argument('docker_run_args', nargs=-1)
@click.argument('docker_image', required=False)
@click.option('--nvidia/--no-nvidia', default=find_executable('nvidia-docker') != None,
              help='Use the nvidia runtime, defaults to nvidia if nvidia-docker is present')
@click.option('--digest', is_flag=True, default=False, help="Output the image digest and exit")
@click.option('--jupyter/--no-jupyter', default=False, help="Run jupyter lab in the container")
@click.option('--dir', default="/app", help="Which directory to mount the code in the container")
@click.option('--no-dir', is_flag=True, help="Don't mount the current directory")
@click.option('--shell', default="/bin/bash", help="The shell to start the container with")
@click.option('--port', default="8888", help="The host port to bind jupyter on")
@click.option('--cmd', help="The command to run in the container")
@click.option('--no-tty', is_flag=True, default=False, help="Run the command without a tty")
@display_error
def docker(ctx, docker_run_args, docker_image, nvidia, digest, jupyter, dir, no_dir, shell, port, cmd, no_tty):
    """W&B docker lets you run your code in a docker image ensuring wandb is configured. It adds the WANDB_DOCKER and WANDB_API_KEY
    environment variables to your container and mounts the current directory in /app by default.  You can pass additional
    args which will be added to `docker run` before the image name is declared, we'll choose a default image for you if
    one isn't passed:

    wandb docker -v /mnt/dataset:/app/data
    wandb docker gcr.io/kubeflow-images-public/tensorflow-1.12.0-notebook-cpu:v0.4.0 --jupyter
    wandb docker wandb/deepo:keras-gpu --no-tty --cmd "python train.py --epochs=5"

    By default we override the entrypoint to check for the existance of wandb and install it if not present.  If you pass the --jupyter
    flag we will ensure jupyter is installed and start jupyter lab on port 8888.  If we detect nvidia-docker on your system we will use
    the nvidia runtime.  If you just want wandb to set environment variable to an existing docker run command, see the wandb docker-run 
    command.
    """
    if not find_executable('docker'):
        raise ClickException(
            "Docker not installed, install it from https://docker.com")
    args = list(docker_run_args)
    image = docker_image or ""
    # remove run for users used to nvidia-docker
    if len(args) > 0 and args[0] == "run":
        args.pop(0)
    if image == "" and len(args) > 0:
        image = args.pop(0)
    # If the user adds docker args without specifying an image (should be rare)
    if not util.docker_image_regex(image.split("@")[0]):
        if image:
            args = args + [image]
        image = wandb.docker.default_image(gpu=nvidia)
        subprocess.call(["docker", "pull", image])
    _, repo_name, tag = wandb.docker.parse(image)

    resolved_image = wandb.docker.image_id(image)
    if resolved_image is None:
        raise ClickException(
            "Couldn't find image locally or in a registry, try running `docker pull %s`" % image)
    if digest:
        sys.stdout.write(resolved_image)
        exit(0)

    existing = wandb.docker.shell(
        ["ps", "-f", "ancestor=%s" % resolved_image, "-q"])
    if existing:
        question = {
            'type': 'confirm',
            'name': 'attach',
            'message': "Found running container with the same image, do you want to attach?",
        }
        result = whaaaaat.prompt([question])
        if result and result['attach']:
            subprocess.call(['docker', 'attach', existing.split("\n")[0]])
            exit(0)
    cwd = os.getcwd()
    command = ['docker', 'run', '-e', 'LANG=C.UTF-8', '-e', 'WANDB_DOCKER=%s' % resolved_image, '--ipc=host',
                '-v', wandb.docker.entrypoint+':/wandb-entrypoint.sh', '--entrypoint', '/wandb-entrypoint.sh']
    if nvidia:
        command.extend(['--runtime', 'nvidia'])
    if not no_dir:
        #TODO: We should default to the working directory if defined
        command.extend(['-v', cwd+":"+dir, '-w', dir])
    if api.api_key:
        command.extend(['-e', 'WANDB_API_KEY=%s' % api.api_key])
    else:
        wandb.termlog("Couldn't find WANDB_API_KEY, run `wandb login` to enable streaming metrics")
    if jupyter:
        command.extend(['-e', 'WANDB_ENSURE_JUPYTER=1', '-p', port+':8888'])
        no_tty = True
        cmd = "jupyter lab --no-browser --ip=0.0.0.0 --allow-root --NotebookApp.token= --notebook-dir %s" % dir
    command.extend(args)
    if no_tty:
        command.extend([image, shell, "-c", cmd])
    else:
        if cmd:
            command.extend(['-e', 'WANDB_COMMAND=%s' % cmd])
        command.extend(['-it', image, shell])
        wandb.termlog("Launching docker container \U0001F6A2")
    subprocess.call(command)



MONKEY_CONTEXT = copy.copy(CONTEXT)
MONKEY_CONTEXT['allow_extra_args'] = True
MONKEY_CONTEXT['ignore_unknown_options'] = True

@cli.command(context_settings=MONKEY_CONTEXT, help="Run any script with wandb", hidden=True)
@click.pass_context
@click.argument('program')
@click.argument('args', nargs=-1)
@display_error
def magic(ctx, program, args):

    def magic_run(cmd, globals, locals):
        try:
            exec(cmd, globals, locals)
        finally:
            pass

    sys.argv[:] = args
    sys.argv.insert(0, program)
    sys.path.insert(0, os.path.dirname(program))
    try:
        with open(program, 'rb') as fp:
            code = compile(fp.read(), program, 'exec')
    except IOError:
        click.echo(click.style("Could not launch program: %s" % program, fg="red"))
        sys.exit(1)
    globs = {
            '__file__': program,
            '__name__': '__main__',
            '__package__': None,
            'wandb_magic_install': magic_install,
        }
    prep = '''
import __main__
__main__.__file__ = "%s"
wandb_magic_install()
''' % program
    magic_run(prep, globs, None)
    magic_run(code, globs, None)


@cli.command(context_settings=CONTEXT, help="Create a sweep")
@click.pass_context
@click.option("--project", "-p", default=None, envvar=env.PROJECT, help="The project of the sweep.")
@click.option("--entity", "-e", default=None, envvar=env.ENTITY, help="The entity scope for the project.")
@click.option('--controller', is_flag=True, default=False, help="Run local controller")
@click.option('--verbose', is_flag=True, default=False, help="Display verbose output")
@click.option('--name', default=False, help="Set sweep name")
@click.option('--program', default=False, help="Set sweep program")
@click.option('--settings', default=False, help="Set sweep settings", hidden=True)
@click.option('--update', default=None, help="Update pending sweep")
@click.argument('config_yaml')
@display_error
def sweep(ctx, project, entity, controller, verbose, name, program, settings, update, config_yaml):
    def _parse_settings(settings):
        """settings could be json or comma seperated assignments."""
        ret = {}
        # TODO(jhr): merge with magic_impl:_parse_magic
        if settings.find('=') > 0:
            for item in settings.split(","):
                kv = item.split("=")
                if len(kv) != 2:
                    wandb.termwarn("Unable to parse sweep settings key value pair", repeat=False)
                ret.update(dict([kv]))
            return ret
        wandb.termwarn("Unable to parse settings parameter", repeat=False)
        return ret

    if api.api_key is None:
        termlog("Login to W&B to use the sweep feature")
        ctx.invoke(login, no_offline=True)

    sweep_obj_id = None
    if update:
        parts = dict(entity=entity, project=project, name=update)
        err = util.parse_sweep_id(parts)
        if err:
            wandb.termerror(err)
            return
        entity = parts.get("entity") or entity
        project = parts.get("project") or project
        sweep_id = parts.get("name") or update
        found = api.sweep(sweep_id, '{}', entity=entity, project=project)
        if not found:
            wandb.termerror('Could not find sweep {}/{}/{}'.format(entity, project, sweep_id))
            return
        sweep_obj_id = found['id']

    wandb.termlog('{} sweep from: {}'.format(
            'Updating' if sweep_obj_id else 'Creating',
            config_yaml))
    try:
        yaml_file = open(config_yaml)
    except (OSError, IOError):
        wandb.termerror('Couldn\'t open sweep file: %s' % config_yaml)
        return
    try:
        config = util.load_yaml(yaml_file)
    except yaml.YAMLError as err:
        wandb.termerror('Error in configuration file: %s' % err)
        return
    if config is None:
        wandb.termerror('Configuration file is empty')
        return

    # Set or override parameters
    if name:
        config["name"] = name
    if program:
        config["program"] = program
    if settings:
        settings = _parse_settings(settings)
        if settings:
            config.setdefault("settings", {})
            config["settings"].update(settings)
    if controller:
        config.setdefault("controller", {})
        config["controller"]["type"] = "local"

    is_local = config.get('controller', {}).get('type') == 'local'
    if is_local:
        tuner = wandb_controller.controller()
        err = tuner._validate(config)
        if err:
            wandb.termerror('Error in sweep file: %s' % err)
            return

    entity = entity or env.get_entity() or config.get('entity')
    project = project or env.get_project() or config.get('project') or util.auto_project_name(
            config.get("program"), api)
    sweep_id = api.upsert_sweep(config, project=project, entity=entity, obj_id=sweep_obj_id)
    wandb.termlog('{} sweep with ID: {}'.format(
            'Updated' if sweep_obj_id else 'Created',
            click.style(sweep_id, fg="yellow")))
    sweep_url = wandb_controller._get_sweep_url(api, sweep_id)
    if sweep_url:
        wandb.termlog("View sweep at: {}".format(
            click.style(sweep_url, underline=True, fg='blue')))

    # reprobe entity and project if it was autodetected by upsert_sweep
    entity = entity or env.get_entity()
    project = project or env.get_project()

    if entity and project:
        sweep_path = "{}/{}/{}".format(entity, project, sweep_id)
    elif project:
        sweep_path = "{}/{}".format(project, sweep_id)
    else:
        sweep_path = sweep_id

    wandb.termlog("Run sweep agent with: {}".format(
            click.style("wandb agent %s" % sweep_path, fg="yellow")))
    if controller:
        wandb.termlog('Starting wandb controller...')
        tuner = wandb_controller.controller(sweep_id)
        tuner.run(verbose=verbose)


@cli.command(context_settings=CONTEXT, help="Run the W&B agent")
@click.pass_context
@click.option("--project", "-p", default=None, envvar=env.PROJECT, help="The project of the sweep.")
@click.option("--entity", "-e", default=None, envvar=env.ENTITY, help="The entity scope for the project.")
@click.option("--count", default=None, type=int, help="The max number of runs for this agent.")
@click.argument('sweep_id')
@display_error
def agent(ctx, project, entity, count, sweep_id):
    if api.api_key is None:
        termlog("Login to W&B to use the sweep agent feature")
        ctx.invoke(login, no_offline=True)

    wandb.termlog('Starting wandb agent 🕵️')
    wandb_agent.run_agent(sweep_id, entity=entity, project=project, count=count)

    # you can send local commands like so:
    # agent_api.command({'type': 'run', 'program': 'train.py',
    #                'args': ['--max_epochs=10']})


@cli.command(context_settings=CONTEXT, help="Run the W&B local sweep controller")
@click.option('--verbose', is_flag=True, default=False, help="Display verbose output")
@click.argument('sweep_id')
@display_error
def controller(verbose, sweep_id):
    click.echo('Starting wandb controller...')
    tuner = wandb_controller.controller(sweep_id)
    tuner.run(verbose=verbose)


if __name__ == "__main__":
    cli()



import collections
import os
import shutil
import threading
import time
from six.moves import queue
import warnings
import tarfile

import wandb
import wandb.util
from wandb.compat import tempfile

def resolve_path(path):
    try:
        from pathlib import Path
        return str(Path(path).resolve())
    except:
        # Pathlib isn't present for python versions earlier than 3.3
        return os.path.realpath(path)

# Get rid of cleanup warnings in Python 2.7.
warnings.filterwarnings('ignore', 'Implicitly cleaning up', RuntimeWarning, 'wandb.compat.tempfile')


# Temporary directory for copies we make of some file types to
# reduce the probability that the file gets changed while we're
# uploading it.
TMP_DIR = tempfile.TemporaryDirectory('wandb')


EventFileBatch = collections.namedtuple(
    'EventFileBatch', ('batch_id', 'file_changed_events',))
EventFileChanged = collections.namedtuple(
    'EventFileChanged', ('path', 'save_name', 'copy'))
EventJobDone = collections.namedtuple('EventJobDone', ('job'))
EventFinish = collections.namedtuple('EventFinish', ())


class UploadJob(threading.Thread):
    def __init__(self, done_queue, progress, api, save_name, path, copy=True):
        """A file upload thread.

        Arguments:
            done_queue: queue.Queue in which to put an EventJobDone event when
                the upload finishes.
            push_function: function(save_name, actual_path) which actually uploads
                the file.
            save_name: string logical location of the file relative to the run
                directory.
            path: actual string path of the file to upload on the filesystem.
            copy: (bool) Whether to copy the file before uploading it. Defaults
                to True because if you try to upload a file while it's being
                rewritten, it's possible that we'll upload something truncated
                or corrupt. Our file-uploading rules are generally designed
                so that that won't happen during normal operation.
        """
        self._done_queue = done_queue
        self._progress = progress
        self._api = api
        self.save_name = save_name
        self.save_path = self.path = path
        self.copy = copy
        self.needs_restart = False
        self.label = save_name
        super(UploadJob, self).__init__()

    def prepare_file(self):
        if self.copy:
            self.save_path = os.path.join(TMP_DIR.name, self.save_name)
            wandb.util.mkdir_exists_ok(os.path.dirname(self.save_path))
            shutil.copy2(self.path, self.save_path)

    def cleanup_file(self):
        if self.copy and os.path.isfile(self.save_path):
            os.remove(self.save_path)

    def run(self):
        try:
            # wandb.termlog('Uploading file: %s' % self.save_name)
            self.prepare_file()
            self.push()
            # wandb.termlog('Done uploading file: %s' % self.save_name)
        finally:
            self.cleanup_file()
            self._done_queue.put(EventJobDone(self))

    def push(self):
        try:
            size = os.path.getsize(self.save_path)
        except OSError:
            size = 0

        self._progress[self.label] = {
            'total': size,
            'uploaded': 0,
            'failed': False
        }
        try:
            with open(self.save_path, 'rb') as f:
                self._api.push(
                    {self.save_name: f},
                    progress=lambda _, t: self.progress(t))
        except Exception as e:
            self._progress[self.label]['uploaded'] = 0
            self._progress[self.label]['failed'] = True
            wandb.util.sentry_exc(e)
            wandb.termerror('Error uploading "{}": {}, {}'.format(
                self.save_name, type(e).__name__, e))

    def progress(self, total_bytes):
        if self.label not in self._progress:
            return
        self._progress[self.label]['uploaded'] = total_bytes

    def restart(self):
        # In the future, this could cancel the current upload and restart it. The logic
        # should go in FilePusher to avoid raciness (it would call job.cancel(),
        # and then restart it).
        self.needs_restart = True


class BatchUploadJob(UploadJob):
    def __init__(self, done_queue, files, api, batch_id, file_changed_events):
        # Copy all files to a temp dir in the correct structure
        tgz_path = os.path.join(TMP_DIR.name, 'batch-%s.tgz' % batch_id)
        # wandb.termlog('Preparing batch: %s' % tgz_path)

        with tarfile.open(tgz_path, 'w:gz') as tar:
            for event in file_changed_events:
                try:
                    tar.add(resolve_path(event.path), arcname=event.save_name)
                except OSError:
                    # Retry once, then show an error and continue
                    time.sleep(0.1)
                    try:
                        tar.add(resolve_path(event.path),
                            arcname=event.save_name)
                    except OSError:
                        wandb.termwarn("Failed to add %s to batch archive." %
                            event.save_name)

        save_name = '___batch_archive_{}.tgz'.format(batch_id)

        super(BatchUploadJob, self).__init__(done_queue, files, api, save_name,
            tgz_path)

        self.label = 'batch_{}'.format(batch_id)
        self.tgz_path = tgz_path

    def cleanup_file(self):
        super(BatchUploadJob, self).cleanup_file()
        # wandb.termlog('Cleaning batch: %s' % self.tgz_path)
        os.unlink(self.tgz_path)


class FileStats(object):
    def __init__(self, save_name, file_path):
        """Tracks file upload progress

        save_name: the file's path in a run. It's an ID of sorts.
        file_path: the local path.
        """
        self._save_name = save_name
        self._file_path = file_path
        self.size = 0
        self.uploaded = 0
        self.failed = False

    def update_size(self):
        try:
            self.size = os.path.getsize(self._file_path)
        except (OSError, IOError):
            pass


class FilePusher(object):
    """Parallel file upload class.

    This manages uploading multiple files in parallel. It will restart a given file's
    upload job if it receives a notification that that file has been modified.
    The finish() method will block until all events have been processed and all
    uploads are complete.
    """

    # After 5 seconds of gathering batched uploads, kick off a batch without
    # waiting any longer.
    BATCH_THRESHOLD_SECS = 3

    # Maximum number of files in any given batch. If there are too many files
    # it can take too long to unpack -- 500 very small files takes GCP about a
    # minute.
    BATCH_MAX_FILES = 100

    # If there are fewer than this many files gathered over a batch threshold, 
    # then just upload them individually.
    BATCH_MIN_FILES = 3

    # If needed you can space out uploads a bit.
    RATE_LIMIT_SECS = 0.1

    def __init__(self, api, max_jobs=6):
        self._file_stats = {}  # stats for all files
        self._progress = {}   # amount uploaded
        self._batch_num = 1  # incrementing counter for archive filenamess

        self._api = api
        self._max_jobs = max_jobs
        self._batch_queue = queue.Queue()
        self._event_queue = queue.Queue()
        self._last_job_started_at = 0
        self._finished = False

        # Thread for processing events and starting upload jobs
        self._process_thread = threading.Thread(target=self._process_body)
        self._process_thread.daemon = True
        self._process_thread.start()

        # Thread for gathering batches and creating a single upload job.
        self._batch_thread = threading.Thread(target=self._batch_body)
        self._batch_thread.daemon = True
        self._batch_thread.start()

        # Indexed by files' `save_name`'s, which are their ID's in the Run.
        self._running_jobs = {}
        self._pending_events = []

    def update_file(self, save_name, file_path):
        if save_name not in self._file_stats:
            self._file_stats[save_name] = FileStats(save_name, file_path)
        self._file_stats[save_name].update_size()

    def rename_file(self, old_save_name, new_save_name, new_path):
        """This only updates the name and path we use to track the file's size
        and upload progress. Doesn't rename it on the back end or make us
        upload from anywhere else.
        """
        if old_save_name in self._file_stats:
            del self._file_stats[old_save_name]
        self.update_file(new_save_name, new_path)

    def update_all_files(self):
        for file_stats in self._file_stats.values():
            file_stats.update_size()

    def print_status(self):
        step = 0
        spinner_states = ['-', '\\', '|', '/']
        stop = False
        while True:
            if not self.is_alive():
                stop = True
            summary = self.summary()
            line = ' %.2fMB of %.2fMB uploaded\r' % (
                summary['uploaded_bytes'] / 1048576.0,
                summary['total_bytes'] / 1048576.0)
            line = spinner_states[step % 4] + line
            if summary['failed_batches']:
                line += ' (%(failed_batches)d failed uploads)'
            step += 1
            wandb.termlog(line, newline=False)
            if stop:
                break
            time.sleep(0.25)
        # clear progress line.
        wandb.termlog(' ' * 79)

    def files(self):
        return self._file_stats.keys()

    def stats(self):
        return self._file_stats

    def summary(self):
        progress_values = list(self._progress.values())
        return {
            'failed_batches': len([f for f in progress_values if f['failed']]),
            'uploaded_bytes': sum(f['uploaded'] for f in progress_values),
            'total_bytes': sum(f['total'] for f in progress_values)
        }

    def _process_body(self):
        # Wait for event in the queue, and process one by one until a
        # finish event is received
        while True:
            event = self._event_queue.get()
            if isinstance(event, EventFinish):
                self._finished = True
                break
            self._process_event(event)

        # After a finish event is received, iterate through the event queue
        # one by one and process all remaining events.
        while True:
            try:
                event = self._event_queue.get(True, 1)
            except queue.Empty:
                event = None
            if event:
                self._process_event(event)
            elif not self._running_jobs:
                # Queue was empty and no jobs left.
                break

    def _batch_body(self):
        # Repeat core loop infinitely until a Finish event is received. Once
        # we've received a finish event, we can terminate the batch thread
        # immediately since it's guaranteed no further file change events
        # will come in.
        finished = False
        while True:
            batch = []
            batch_started_at = time.time()
            batch_end_at = batch_started_at + self.BATCH_THRESHOLD_SECS
            while time.time() < batch_end_at and len(batch) < self.BATCH_MAX_FILES:
                # Get the latest event
                try:
                    wait_secs = batch_end_at - time.time()
                    event = self._batch_queue.get(timeout=wait_secs)
                except queue.Empty:
                    # If nothing is available in the batch by the timeout
                    # wrap up and send the current batch immediately.
                    break
                # If it's a finish, stop waiting and send the current batch
                # immediately.
                if isinstance(event, EventFinish):
                    finished = True
                    break
                # Otherwise, it's file changed, so add it to the pending batch.
                batch.append(event)

            if batch:
                if len(batch) <= self.BATCH_MIN_FILES:
                    # If less than the minimum files are found, just upload
                    # them individually.
                    for event in batch:
                        self._event_queue.put(event)
                else:
                    # Otherwise, send all the files as a batch.
                    new_batch_id = str(self._batch_num)
                    self._event_queue.put(EventFileBatch(new_batch_id, batch))
                    self._batch_num += 1
            
            # And stop the infinite loop if we've finished
            if finished:
                break

    def _process_event(self, event):
        if isinstance(event, EventJobDone):
            job = event.job
            job.join()
            self._running_jobs.pop(job.label)
            if job.needs_restart:
                #wandb.termlog('File changed while uploading, restarting: %s' % event.job.save_name)
                self._start_or_restart_event_job(event)
            elif self._pending_events:
                event = self._pending_events.pop()
                self._start_or_restart_event_job(event)
            return

        # If it gets here, must be a FileChanged or FileBatch event
        if len(self._running_jobs) == self._max_jobs:
            self._pending_events.append(event)
            return
            
        # Start now if we have capacity
        self._start_or_restart_event_job(event)

    def _label_for_event(self, event):
        if isinstance(event, EventFileChanged):
            return event.save_name
        if isinstance(event, EventFileBatch):
            return 'batch_{}'.format(event.batch_id)
        return None

    def _start_or_restart_event_job(self, event):
        label = self._label_for_event(event)
        if not label:
            return

        # Restart if in running jobs
        if label in self._running_jobs:
            self._running_jobs[label].restart()
            return

        # Rate limit if it's too fast to prevent overloading the server
        elapsed_since_last = time.time() - self._last_job_started_at
        if elapsed_since_last < self.RATE_LIMIT_SECS:
            time.sleep(self.RATE_LIMIT_SECS - elapsed_since_last)
            self._start_or_restart_event_job(event)
            return

        # Or start
        self._last_job_started_at = time.time()
        self._running_jobs[label] = self._start_event_job(label, event)

    def _start_event_job(self, label, event):
        if isinstance(event, EventFileChanged):
            return self._start_single_job(event.save_name, event.path,
                event.copy)

        if isinstance(event, EventFileBatch):
            return self._start_batch_job(event.batch_id,
                event.file_changed_events)

    def _start_single_job(self, save_name, path, copy):
        # wandb.termlog("Starting individual upload: %s" % save_name)
        job = UploadJob(self._event_queue, self._progress, self._api, save_name, path, copy)
        job.start()
        return job

    def _start_batch_job(self, batch_id, file_changed_events):
        # wandb.termlog("Starting batch %s (%d files)" % (batch_id,
        #     len(file_changed_events)))
        job = BatchUploadJob(self._event_queue, self._progress, self._api,
            batch_id, file_changed_events)
        job.start()
        return job

    def should_batch(self, file_change_event):
        """
        Whether a file gets batched depends on file size. Anything above
        1MB should be handled individually.
        """
        return os.path.getsize(file_change_event.path) < 1000000

    def file_changed(self, save_name, path, copy=True):
        """Tell the file pusher that a file's changed and should be uploaded.

        Arguments:
            save_name: string logical location of the file relative to the run
                directory.
            path: actual string path of the file to upload on the filesystem.
            copy: (bool) Whether to copy the file before uploading it. Defaults
                to True because if you try to upload a file while it's being
                rewritten, it's possible that we'll upload something truncated
                or corrupt. Our file-uploading rules are generally designed
                so that that won't happen during normal operation.
        """
        # Tests in linux were failing because wandb-events.jsonl didn't exist
        if not os.path.exists(path):
            return
        if os.path.getsize(path) == 0:
            return

        event = EventFileChanged(path, save_name, copy)

        if self.should_batch(event):
            self._batch_queue.put(event)
        else:
            self._event_queue.put(event)

    def finish(self):
        self._event_queue.put(EventFinish())
        self._batch_queue.put(EventFinish())

    def shutdown(self):
        self.finish()
        self._batch_thread.join()
        self._process_thread.join()

    def is_alive(self):
        return self._process_thread.is_alive() or self._batch_thread.is_alive()
